! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module detect_io

!-------------------------------------------------------------------------------
! Module containing input/output routines for eROSITA detection  tasks.
! Author: Georg Lamer , AIP         date:  19-MAY-2010  
! update: GLamer                    date:  26-MAY-2010 : use numkind
! update: GLamer                    date:  18-02-2011  : add max numbers of images
!                                                        add sort_files 
! update: GLamer                    date:  02-05-2011  : add subroutines addval,addwitherr,
!                                                        average, w_average, dw_average, hrcalc
!                                                        data types  ermldet_param_type, erbox_param_type
! update: GLamer                    date:  16-09-2011  : Change column names in mllist: det_ml -> like
!                                                        ext_ml -> ext_like , ml_id_src -> id_src
!                                                        in boxlist: box_id_src -> id_src
! update: GLamer                    date:  25-05-2012  : 'character function lowcase' -> character(len=clong) function lowcase
!                                                        (required by gfortran 4.6)
!                                                         Changed to Fortran 2003 style array constructors for character arrays 
!                                                         in fput_boxlist and fput_mllist 
! update: GLamer                    date:  25-05-2012  :  Changes to FORMAT_MLLIST (fixed RATE and FLUX columns)
! update: GLamer                    date:  15-11-2012  :  Changes to FORMAT_BOXLIST (allocate(srcbuff) )
! update: GLamer                    date:  05-06-2013  :  maxsou: 100000 -> 300000

 
!                                  
!
!
!     
! subroutines:
!  fget_mllist
!  fput_mllist
!  fget_boxlist
!  fput_boxlist
!  filter_srclist
!  format_boxlist
!  format_mllist
!  init_mllist1d
!  init_mllist2d
!  init_mllist3d
!  init_boxlist1d
!  init_boxlist2d
!  init_boxlist3d

!  sort_files
!  fget_multi
!  addval
!  addwitherr
!  average
!  w_average
!  dw_averag
!  hrcalc
!  gammq
!  gammp
!  gammln
!  
! functions:
!   ixb
!   iyb
!
! data types:
!   ermldet_param_type
!   erbox_param_type
!   mlrow_type
!   mllist_type
!   boxrow_type
!   srcrow_type
!   boxlist_type
!   boxlist2d_type
!   boxlist3d_type 
!-------------------------------------------------------------------------------

  use param
  use fitsimage
  use cfitsio_ifaces
  use numkind
  use astrolib
  use, intrinsic :: iso_fortran_env
 ! use, intrinsic :: ieee_arithmetic


  ! avoid "use cal", define private versions of deg2rad, rad2deg
  real(kind=real64),parameter,private       ::  deg2rad=1.74532925199D-02, rad2deg=57.29577951322  



  public 


!! moved to fitsimage_mod.f90: 
!    integer(kind=int32)             :: i4null = -999999
!    real(kind=real32)               :: r4null = -9.999E9
!    real(kind=real64)               :: r8null = -9.999D9 
!    character(len=1)                :: cnull= ''

!    integer(kind=int32),parameter         :: max_cam = 8 , num_tel=7   
! !   character(len=8), dimension(max_cam)  :: cam_names = (/'TEL1','TEL2','TEL3', &
! !                                            'TEL4','TEL5','TEL6','TEL7','TEL0'/)
! !   ! For testing purposes , use instruments SIM1 .. SIM7
!    character(len=8), dimension(max_cam)  :: cam_names = (/'SIM1','SIM2','SIM3', &
!                                             'SIM4','SIM5','SIM6','SIM7','SIM0'/)
!    -------------------------------------------------------------------------------
   
    
    ! global parameters:
    integer(kind=int32),parameter  ::  max_exp  = 60
    ! set to 10 for testing:
    !integer(kind=int32),parameter  ::  max_bands= 5 
     integer(kind=int32),parameter  ::  max_bands= 10 

    integer(kind=int32),parameter  ::  max_ima  = max_exp*max_bands 
    integer(kind=int32),parameter  ::  max_sou  = 1000000
    integer(kind=int32),parameter  ::  max_hr   = 3 
    integer(kind=int32),parameter  ::  max_nmul = 8  
    integer(kind=int32),parameter  ::  max_mask = 256    
    integer(kind=int32),parameter  ::  maskthreshold = 0.
    integer(kind=int32),parameter  ::  psfthreshold = 0.15        
          
    integer, parameter :: max_fit = 10
    integer, parameter :: max_par = max_fit*max_ima+max_fit*3
   
    ! global variables:
    
    integer(kind=int32)                                 :: num_exp, num_cam, nim
    integer(kind=int32), dimension(max_ima)             :: mask_ip,expip,instip,bandip
    integer(kind=int32), dimension(max_exp, max_bands)  :: ip
    integer(kind=int32), dimension(max_exp)             :: num_image
!    character(len=21), dimension(max_ima) ,public       :: instrume
    character(len=128), dimension(max_ima) ,public       :: instrume
    ! variables relatec to compressed (memory saving) image reading
    ! use global variables for performance and simplicity:
    integer(kind=int32)                                 :: ximbuf,yimbuf
    integer(kind=int32), dimension(max_exp)             :: ixb0,iyb0
    logical(kind=log32)                                 :: compress_flag

    ! test variables:
    integer(kind=int32)   :: fcounter, pcounter, mcounter, lcounter, kcounter, cclock1, cclock2, cclock0

    
    !------------------------------------------------------------------------------ 
    ! Description of global variables/parameters:
    ! -------------------------------------------
    !
    ! MAX_CAM:   Total number of "hardware" instruments, i.e.
    !            instruments (or combinations of them) with  differing response 
    !            characteristics.
    !            Source parameters from each instrument will be written  
    !            to separate rows of source lists. 
    ! CAM_NAMES: Names of hardware instruments as in header keyword 'instrume'.
    ! MAX_EXP  : Maximum allowed number of exposures (XMM_SAS: max_ins).
    ! MAX_BANDS: Maximum number of energy bands per exposure.
    ! MAX_IMA  : Maximum number of input images. 
    ! MAX_SOU  : Size of internal source buffer.
    !
    ! NUM_EXP  : Number of exposures sum over number of instruments (num_cam)
    !             * exposures / per instrument (e.g. raster pointings).
    !            (XMM_SAS: num_instr).
    ! NUM_CAM  : Number of instruments used.  
    ! IP       : 2D array of pointers to input images, values                        
    !            refer to order of image in input list.
    ! EXPIP    : Array to camera indices, values refer to position
    !            of camera in string array CAM_NAMES.
    ! XIMBUF
    ! YIMBUF   : dimensions of image buffers in case of memory saving reading 
    ! IXB0,IYB0: x,y - offsets for each image  (memory saving reading)
    !  
    !------------------------------------------------------------------------------


  type events_type
!    real(kind=4)    :: xs, ys, pixsiz, exp, bg
!    integer(kind=4) :: n, ipt, ins, offax, posang
    integer(kind=4)  :: n, ix, iy, ipt, id
  end type events_type
 

   type mask_type
      integer(kind=int8), pointer, dimension(:,:,:)  :: mask
      integer(kind=int32)                            :: ix_min, ix_max, iy_min, iy_max
      integer(kind=int32)                            :: nx, ny, nev, npix, nsrc
      real(kind=real32),dimension(max_fit)           :: srcreg_x,srcreg_y,srcreg_r
   end type mask_type


  type ermldet_param_type

    
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  boxlistset
    character(len=clong)                    ::  mllistset
    character(len=clong),dimension(max_ima) ::  imagesets
    integer(kind=int32)                     ::  n_imagesets
    character(len=clong),dimension(max_ima) ::  expimagesets
    integer(kind=int32)                     ::  n_expimagesets
    character(len=clong),dimension(max_ima) ::  detmasksets
    integer(kind=int32)                     ::  n_detmasksets
    character(len=clong),dimension(max_ima) ::  bkgimagesets
    integer(kind=int32)                     ::  n_bkgimagesets
    character(len=clong),dimension(max_ima) ::  srcimagesets
    integer(kind=int32)                     ::  n_srcimagesets
    real(kind=real32),dimension(max_ima)    ::  emin
    integer(kind=int32)                     ::  n_emin
    real(kind=real32),dimension(max_ima)    ::  emax
    integer(kind=int32)                     ::  n_emax
    real(kind=real32),dimension(max_ima)    ::  ecf
    integer(kind=int32)                     ::  n_ecf
    integer(kind=int32),dimension(2*max_hr) ::  hrdef
    integer(kind=int32)                     ::  n_hrdef           
    real(kind=real32)                       ::  likemin
    real(kind=real32)                       ::  extlikemin
    real(kind=real32),dimension(2)          ::  extlike_slope
    integer(kind=int32)                     ::  n_extlike_slope
    real(kind=real32)                       ::  extmin
    real(kind=real32)                       ::  extmax
    character(len=80)                       ::  extentmodel
    integer(kind=int32)                     ::  nmaxfit
    integer(kind=int32)                     ::  nmulsou
    logical(kind=4)                         ::  thres_flag
    real(kind=real32)                       ::  thres_val
    character(len=80)                       ::  thres_col
    real(kind=real32)                       ::  multrad
    real(kind=real32)                       ::  cutrad  
    logical(kind=4)                         ::  shapelet_flag
    logical(kind=4)                         ::  photon_flag
    logical(kind=4)                         ::  fitext_flag
    logical(kind=4)                         ::  fitpos_flag
    logical(kind=4)                         ::  twostage_flag
    logical(kind=4)                         ::  compress_flag    
    logical(kind=4)                         ::  srcima_flag          
    logical(kind=4)                         ::  expima_flag
    logical(kind=4)                         ::  detmask_flag
    logical(kind=4)                         ::  sixte_flag 
    ! this is currently an internal only parameter (read extent values from input, keep fixed):
    logical(kind=4)                         ::  fixed_ext_flag 
 
 

  end type  ermldet_param_type



  type erbox_param_type

    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong),dimension(max_ima) ::  imagesets
    integer(kind=int32)                     ::  n_imagesets
    character(len=clong)                    ::  boxlistset
    character(len=clong),dimension(max_ima) ::  expimagesets
    integer(kind=int32)                     ::  n_expimagesets
    character(len=clong),dimension(max_ima) ::  detmasksets
    integer(kind=int32)                     ::  n_detmasksets
    character(len=clong),dimension(max_ima) ::  bkgimagesets
    integer(kind=int32)                     ::  n_bkgimagesets
    real(kind=real32),dimension(max_ima)    ::  emin
    integer(kind=int32)                     ::  n_emin
    real(kind=real32),dimension(max_ima)    ::  emax
    integer(kind=int32)                     ::  n_emax
    real(kind=real32),dimension(max_ima)    ::  ecf
    integer(kind=int32)                     ::  n_ecf
    integer(kind=int32),dimension(2*max_hr) ::  hrdef
    integer(kind=int32)                     ::  n_hrdef
    integer(kind=int32)                     ::  nruns
    real(kind=real32)                       ::  likemin
    integer(kind=int32)                     ::  boxsize
    logical(kind=4)                         ::  compress_flag
    logical(kind=4)                         ::  bkgima_flag
    logical(kind=4)                         ::  expima_flag
    logical(kind=4)                         ::  detmask_flag
  end type  erbox_param_type
 

 
  type ermask_param_type
    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  expimage
    character(len=clong)                    ::  detmask
    real(kind=real32)                       ::  threshold1
    real(kind=real32)                       ::  threshold2
    logical(kind=4)                         ::  regionfile_flag
    character(len=clong)                    ::  regionfile
  end type  ermask_param_type
 

   

  type mlrow_type
    !sequence
    integer(kind=int32) :: id_src
    integer(kind=int32) :: box_id_src
    integer(kind=int32) :: id_inst
    integer(kind=int32) :: id_band
    integer(kind=int32) :: id_cluster
    real(kind=real32)   :: scts 
    real(kind=real32)    :: scts_err
    real(kind=real32)    :: x_ima
    real(kind=real32)    :: x_ima_err       
    real(kind=real32)    :: y_ima
    real(kind=real32)    :: y_ima_err  
    real(kind=real32)    :: ext   
    real(kind=real32)    :: ext_err
    real(kind=real32)    :: like   
    real(kind=real32)    :: ext_like 
    real(kind=real32)    :: bg_map   
    real(kind=real32)    :: exp_map
    real(kind=real32)    :: flux
    real(kind=real32)    :: flux_err       
    real(kind=real32)    :: rate
    real(kind=real32)    :: rate_err   
    real(kind=real64)    :: ra
    real(kind=real64)    :: dec       
    real(kind=real32)    :: radec_err         
    real(kind=real64)    :: lII       
    real(kind=real64)    :: bII         
    integer(kind=int32)  :: rawx   
    integer(kind=int32)  :: rawy
    real(kind=real32)    :: offax
    integer(kind=int32)  :: ccdnr
    real(kind=real32)    :: hr1      
    real(kind=real32)    :: hr1_err         
    real(kind=real32)    :: hr2      
    real(kind=real32)    :: hr2_err         
    real(kind=real32)    :: hr3       
    real(kind=real32)    :: hr3_err 
!    real(kind=real32)    :: hr4       
!    real(kind=real32)    :: hr4_err 
    real(kind=real32)    :: cutrad
    real(kind=real32)    :: maskfrac
    real(kind=real32)    :: eef   
    real(kind=real32)    :: vignetting  
    real(kind=real32)    :: ontime
    real(kind=real32)    :: dist_nn 
!    character(len=12)    :: flag 
  end type mlrow_type  


  type srcrow_type
    !sequence
    integer(kind=int32) :: id_src
    integer(kind=int32) :: box_id_src
    integer(kind=int32) :: id_inst
    integer(kind=int32) :: id_band
    integer(kind=int32) :: id_cluster
    real(kind=real32)   :: scts 
    real(kind=real32)    :: scts_err
    real(kind=real32)    :: box_cts   
    real(kind=real32)    :: x_ima
    real(kind=real32)    :: x_ima_err       
    real(kind=real32)    :: y_ima
    real(kind=real32)    :: y_ima_err  
    real(kind=real32)    :: ext   
    real(kind=real32)    :: ext_err
    real(kind=real32)    :: like   
    real(kind=real32)    :: ext_like 
    real(kind=real32)    :: bg_map   
    real(kind=real32)    :: bg_raw
    real(kind=real32)    :: exp_map
    real(kind=real32)    :: flux
    real(kind=real32)    :: flux_err       
    real(kind=real32)    :: rate
    real(kind=real32)    :: rate_err   
    real(kind=real64)    :: ra
    real(kind=real64)    :: dec       
    real(kind=real32)    :: radec_err         
    real(kind=real64)    :: lII       
    real(kind=real64)    :: bII         
    integer(kind=int32)  :: rawx   
    integer(kind=int32)  :: rawy
    real(kind=real32)    :: offax
    integer(kind=int32)  :: ccdnr
    real(kind=real32)    :: hr1      
    real(kind=real32)    :: hr1_err         
    real(kind=real32)    :: hr2      
    real(kind=real32)    :: hr2_err         
    real(kind=real32)    :: hr3       
    real(kind=real32)    :: hr3_err 
!    real(kind=real32)    :: hr4       
!    real(kind=real32)    :: hr4_err 
    real(kind=real32)    :: cutrad
    real(kind=real32)    :: maskfrac
    real(kind=real32)    :: vignetting
    integer(kind=int32)  :: box_size         
    real(kind=real32)    :: eef         
    real(kind=real32)    :: ontime
    real(kind=real32)    :: dist_nn 
!    character(len=12)    :: flag 
    integer(kind=int8)  :: owner   
    ! internal parameters, not written to
    ! fits table:
    integer(kind=int32)  :: iwind 
    integer(kind=int32)  :: maskzero
    integer(kind=int32)  :: nzero  
    integer(kind=int32)  :: ndof 
    real(kind=real32)    :: snr
    ! ermldet internal:  
    logical(kind=4)      :: cluflag 
    real(kind=real32)    :: scutrad
    real(kind=real32)    :: ecutrad
!    ! AGE HAS ADDED THOSE 
!    ! THESE ARE APETOOL RELATED 
!    ! PARAMETERS
    integer(kind=int32)  :: ape_cts 
    real(kind=real32)    :: ape_bkg
    real(kind=real32)    :: ape_exp   
    real(kind=real32)    :: ape_eef
    real(kind=real32)    :: ape_radius       
    real(kind=real32)    :: ape_pois

  end type srcrow_type  

 
  type mllist_type
    !sequence
    type(shapehdr_type)                      :: shapehdr
    type(stdhdr_type)                        :: stdhdr
!    type(mlrow_type), pointer, dimension(:)  :: data  => NULL()
    type(srcrow_type), pointer, dimension(:)  :: data  => NULL()
    integer(kind=int32)                       :: ncol=49


    ! removed HR4, HR4_ERR,FLAG
    character(len=16), dimension(49) :: cname = [ character(len=16) :: &
            'ID_SRC'   ,'BOX_ID_SRC' ,'ID_INST' ,'ID_BAND'   ,'ID_CLUSTER', &
            'ML_CTS'     ,'ML_CTS_ERR'   ,'X_IMA'   ,'X_IMA_ERR' ,'Y_IMA', &
            'Y_IMA_ERR','EXT'        ,'EXT_ERR' ,'DET_LIKE' ,'EXT_LIKE', &
            'ML_BKG'   ,'ML_EXP'    ,'ML_FLUX'  ,'ML_FLUX_ERR ' ,'ML_RATE', &
            'ML_RATE_ERR' ,'RA'      ,'DEC'     ,'RADEC_ERR' ,'LII', &
            'BII'      ,'RAWX'       ,'RAWY'    ,'OFFAX'     ,'TM_NR', &
            'HR1 '     ,'HR1_ERR '   ,'HR2'     ,'HR2_ERR'   ,'HR3', &
            'HR3_ERR'  ,'ML_RADIUS'    ,'MASKFRAC', &
            'ML_EEF'   ,'VIGNET' ,'ONTIME'  ,'DIST_NN'   ,'OWNER',   &
            'APE_CTS'  ,'APE_BKG'    ,'APE_EXP'  ,'APE_EEF'   ,'APE_RADIUS', &
            'APE_POIS' ]

    character(len=4) , dimension(49) :: cform=[ character(len=4)  ::   &
            '1J  ','1J  ','1J  ','1J  ','1J  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
            '1D  ','1D  ','1E  ','1D  ','1D  ','1J  ','1J  ','1E  ','1J  ','1E  ','1E  ','1E  ', & 
            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1B  ', &
            '1J  ','1E  ','1E  ','1E  ','1E  ','1E  ']


   character(len=16), dimension(49) :: cunit=[ character(len=16) :: &
            '','','','','','','','pixel','pixel','pixel','pixel','arcsec', &
            'arcsec','','','count /arcmin**2','s','erg /cm**2 /s','erg /cm**2 /s','count /s','count /s','deg','deg', &
            'arcsec','deg','deg','pixel','pixel','arcmin','','','','', &
            '','','','','','','','s','', &
            '','','','','','arcsec','']   
 

  end type mllist_type


  type boxrow_type
    !sequence
    integer(kind=int32)  :: id_src
    integer(kind=int32)  :: id_inst
    integer(kind=int32)  :: id_band  
    real(kind=real32)    :: scts  
    real(kind=real32)    :: scts_err
    real(kind=real32)    :: box_cts       
    real(kind=real32)    :: x_ima
    real(kind=real32)    :: x_ima_err       
    real(kind=real32)    :: y_ima
    real(kind=real32)    :: y_ima_err       
    real(kind=real32)    :: like
    real(kind=real32)    :: bg_map         
    real(kind=real32)    :: bg_raw
    real(kind=real32)    :: exp_map       
    real(kind=real32)    :: flux
    real(kind=real32)    :: flux_err       
    real(kind=real32)    :: rate
    real(kind=real32)    :: rate_err   
    real(kind=real64)    :: ra
    real(kind=real64)    :: dec       
    real(kind=real32)    :: radec_err         
    real(kind=real64)    :: lII       
    real(kind=real64)    :: bII         
    real(kind=real32)    :: hr1      
    real(kind=real32)    :: hr1_err         
    real(kind=real32)    :: hr2      
    real(kind=real32)    :: hr2_err         
    real(kind=real32)    :: hr3       
    real(kind=real32)    :: hr3_err         
    real(kind=real32)    :: vignetting       
    integer(kind=int32)  :: box_size         
    real(kind=real32)    :: eef       
    real(kind=real32)    :: dist_nn 
    ! internal parameters, not written to
    ! fits table:  
    integer(kind=int32)  :: iwind    
    integer(kind=int32)  :: maskzero
    integer(kind=int32)  :: nzero        
    integer(kind=int32)  :: ndof 
    ! ermldet internal:  
    logical(kind=4)      :: cluflag 
    real(kind=real32)    :: scutrad
    real(kind=real32)    :: ecutrad
   end type boxrow_type     
    
   
  type boxlist_type
    !sequence
    type(shapehdr_type)                      :: shapehdr
    type(stdhdr_type)                        :: stdhdr
!    type(boxrow_type), pointer, dimension(:) :: data  => NULL() 
    type(srcrow_type), pointer, dimension(:) :: data  => NULL() 
    integer(kind=int32)                      :: ncol=34
    logical(kind=4)                          :: has_extentcol

 

    character(len=16), dimension(34) :: cname = [ character(len=16) :: &
                            'ID_SRC','ID_INST','ID_BAND','SCTS','SCTS_ERR','BOX_CTS','X_IMA','X_IMA_ERR',  &
                            'Y_IMA','Y_IMA_ERR','LIKE','BG_MAP','BG_RAW','EXP_MAP','FLUX','FLUX_ERR',  &
                            'RATE','RATE_ERR','RA','DEC','RADEC_ERR','LII','BII',  &
                            'HR1','HR1_ERR','HR2','HR2_ER','HR3','HR3_ERR','VIGNETTING',  &
                            'BOX_SIZE','EEF','DIST_NN','EXT']  
    
    character(len=4) , dimension(34) :: cform=[ character(len=4)  ::   &
                         '1J  ','1J  ','1J  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
                         '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', & 
                         '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1J  ','1E  ','1E  ','1E  ']

    character(len=16), dimension(34) :: cunit=[ character(len=16) :: &
                                        '','','','','','','','','','','','','', &
                                        '','','','','','','','','','',          &
                                        '','','','','','','','','','','']
 
  end type boxlist_type

 
  type boxlist2d_type
    !sequence
    type(shapehdr_type)                        :: shapehdr
    type(stdhdr_type)                          :: stdhdr
!    type(boxrow_type), pointer, dimension(:,:) :: data => NULL() 
    type(srcrow_type), pointer, dimension(:,:) :: data => NULL() 
  end type boxlist2d_type


  type boxlist3d_type
    !sequence
    type(shapehdr_type)                          :: shapehdr
    type(stdhdr_type)                            :: stdhdr
!    type(boxrow_type), pointer, dimension(:,:,:) :: data   => NULL() 
    type(srcrow_type), pointer, dimension(:,:,:) :: data   => NULL() 
  end type boxlist3d_type



contains  

!-------------------------------------------------------------------------------

  subroutine fget_mllist(filename,mllist,ierr)


    implicit none
  
    type(mllist_type)                :: mllist
    character(len=*), intent(in)     :: filename
    character(len=7)                 :: extname='SRCLIST'
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(mllist%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_MLLIST'
    character(len=80)                :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
         

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening mllist')
      else  
!        allocate(mllist)
        call fread_stdhdr(unit,mllist%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,extname,0,status)
        if (status /= 0) then
          ierr = 3
          errstr = 'error moving to extension '//extname 
          call WRERR(rtname,ierr,errstr)
        else 
    
          call fread_shapehdr(unit,mllist%shapehdr,status)
          nrows=mllist%shapehdr%naxis2
          allocate(mllist%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          do icol=1,mllist%ncol
            call FTGCNO(unit,.false.,mllist%cname(icol),colnum(icol),status)
            ! enable reading of old style column names  (16/03/2018)

            if (mllist%cname(icol) == "ML_CTS" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"SCTS",colnum(icol),status)
            endif

 
            if (mllist%cname(icol) == "ML_CTS_ERR" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"SCTS_ERR",colnum(icol),status)
            endif

            if (mllist%cname(icol) == "DET_LIKE" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"LIKE",colnum(icol),status)
            endif

 
            if (mllist%cname(icol) == "ML_BKG" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"BG_MAP",colnum(icol),status)
            endif
 
            if (mllist%cname(icol) == "ML_EXP" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"EXP_MAP",colnum(icol),status)
            endif
 
            if (mllist%cname(icol) == "ML_FLUX" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"FLUX",colnum(icol),status)
            endif

            if (mllist%cname(icol) == "ML_FLUX_ERR" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"FLUX_ERR",colnum(icol),status)
            endif

            if (mllist%cname(icol) == "ML_RATE" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"RATE",colnum(icol),status)
            endif

            if (mllist%cname(icol) == "ML_RATE_ERR" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"RATE_ERR",colnum(icol),status)
            endif

            if (mllist%cname(icol) == "TM_NR" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"CCDNR",colnum(icol),status)
            endif
            if (mllist%cname(icol) == "ML_RADIUS" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"CUTRAD",colnum(icol),status)
            endif
            if (mllist%cname(icol) == "ML_EEF" .and. status == 219) then
              status=0
              call FTGCNO(unit,.false.,"EEF",colnum(icol),status)
            endif



            if (status > 0) then
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//mllist%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif


            status=0
          enddo



          do while (nrows > irow)  
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            icol=1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%id_src  ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%box_id_src ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%id_inst    ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%id_band    ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%id_cluster ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%scts       ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                         
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%scts_err   ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%x_ima      ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                            
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%x_ima_err  ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                      
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%y_ima     ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                            
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%y_ima_err ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ext       ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ext_err   ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%like      ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ext_like  ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%bg_map     ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%exp_map    ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%flux       ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                             
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%flux_err   ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                        
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%rate       ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                          
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%rate_err   ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                       
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,mllist%data(i1:i2)%ra         ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                             
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,mllist%data(i1:i2)%dec        ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                         
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%radec_err  ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)                      
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,mllist%data(i1:i2)%lII        ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                           
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,mllist%data(i1:i2)%bII        ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                           
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%rawx       ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%rawy       ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%offax  ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%ccdnr      ,fnull,status)
            icol=icol+1
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr1        ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)                           
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr1_err    ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)                       
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr2        ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                           
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr2_ERR    ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                      
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr3        ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)                           
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr3_err    ,fnull,status)
            icol=icol+1 
            fstat = max(fstat,status)                       
!            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr4        ,fnull,status) 
!            icol=icol+1
!            fstat = max(fstat,status)                           
!            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%hr4_err    ,fnull,status)
!            icol=icol+1 
!            fstat = max(fstat,status)    
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ecutrad     ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)     
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%maskfrac   ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)    
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%eef        ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)    
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%vignetting ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)    
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ontime     ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status) 
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%dist_nn     ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status) 
            if (colnum(icol) > 0) call  FTGCVB(unit,colnum(icol),i1,1,rowbuf,i1null,mllist%data(i1:i2)%owner         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            ! AGE HAS ADDED APETOOL RELATED PARAMETERS
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,mllist%data(i1:i2)%ape_cts         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ape_bkg         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ape_exp         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ape_eef         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ape_radius         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,mllist%data(i1:i2)%ape_pois         ,fnull,status) 
            icol=icol+1
            fstat = max(fstat,status)  
            ! AGE END EDITS
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif



   

  end subroutine fget_mllist

!-------------------------------------------------------------------------------


  subroutine fput_mllist(filename,mllist,ierr)

    implicit none
 
  
    type(mllist_type)                  :: mllist
    character(len=*), intent(in)       :: filename
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
     integer(kind=int32)               :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: nhdu, varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    integer(kind=int32), parameter     :: ncol=46
!    character(len=16), dimension(ncol) :: ttype,tunit
!    character(len=4), dimension(ncol)  :: tform
    character(len=16)                  :: extname
    character(len=*),parameter         :: rtname = 'FPUT_MLLIST'
    character(len=80)                  :: errstr

    extname='SRCLIST'

   ! NOTE: All elements of character array need to be assigned by strings of equal length: 

!    ttype=[ character(len=16) :: 'ID_SRC          ','BOX_ID_SRC      ','ID_INST         ','ID_BAND         ','ID_CLUSTER      ', & 
!            'SCTS            ','SCTS_ERR        ','X_IMA           ','X_IMA_ERR       ','Y_IMA           ', &
!            'Y_IMA_ERR       ','EXT             ','EXT_ERR         ','LIKE            ','EXT_LIKE        ', &
!            'BG_MAP          ','EXP_MAP         ','FLUX            ','FLUX_ERR        ','RATE            ', &
!            'RATE_ERR        ','RA              ','DEC             ','RADEC_ERR       ','LII             ', &
!            'BII             ','RAWX            ','RAWY            ','OFFAX           ','CCDNR           ', &
!            'HR1             ','HR1_ERR         ','HR2             ','HR2_ERR         ','HR3             ', &
!            'HR3_ERR         ','HR4             ','HR4_ERR         ','CUTRAD          ','MASKFRAC        ', &
!            'EEF             ','VIGNETTING      ','ONTIME          ','DIST_NN         ','FLAG            ',  'OWNER           ']
 

!    tform=[ character(len=4) :: '1J  ','1J  ','1J  ','1J  ','1J  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
!            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
!            '1D  ','1D  ','1E  ','1D  ','1D  ','1J  ','1J  ','1E  ','1J  ','1E  ','1E  ','1E  ', & 
!            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','12A ','1I  ']

    
!    tunit=[ character(len=16) :: '','','','','','','','','','','', &        
!            '','','','','','','','','','','', &
!            '','','','','','','','','','','', &
!            '','','','','','','','','','','','','']   
 

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=mllist%shapehdr%naxis2
    


    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    call FTINIT(unit,filename,blocksize,status)
    fstat=max(fstat,status)
    if (fstat /= 0) then    
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else    
      status=0
      simple   = .true.
      bitpix   = 16
      naxis      = 0
      naxes(1) = 0
      !write FITS standard header:   
      call FTPHPS(unit,bitpix,naxis,naxes,status) 
      fstat=max(fstat,status)  
      ! write eROSITA standard header to primary extension :
      ! now obsolete?? (29/04/2015)
      call fwrite_stdhdr(unit,mllist%stdhdr,status)
      ! create new empty HDU:
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  
      ! write header for binary table:
      call FTPHBN(unit,nrows,mllist%ncol,mllist%cname,mllist%cform,mllist%cunit,extname,varidat,status)
      fstat=max(fstat,status)  
      ! put standard header into table extension (29/04/2015): 
      call fwrite_stdhdr(unit,mllist%stdhdr,status)
      fstat=max(fstat,status)  


      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
     
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 

        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2        
          icol=1
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%id_src  ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%box_id_src ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%id_inst    ,status)
          icol=icol+1
          fstat = max(fstat,status)          
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%id_band    ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%id_cluster ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%scts       ,status)
          icol=icol+1
          fstat = max(fstat,status)                     
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%scts_err   ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%x_ima      ,status) 
          icol=icol+1
          fstat = max(fstat,status)                        
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%x_ima_err  ,status)
          icol=icol+1
          fstat = max(fstat,status)                    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%y_ima     ,status) 
          icol=icol+1
          fstat = max(fstat,status)                        
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%y_ima_err ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ext       ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ext_err   ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%like    ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ext_like    ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%bg_map    ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%exp_map   ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%flux       ,status)
          icol=icol+1
          fstat = max(fstat,status)                         
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%flux_err   ,status)
          icol=icol+1
          fstat = max(fstat,status)                    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%rate       ,status) 
          icol=icol+1
          fstat = max(fstat,status)                      
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%rate_err   ,status)
          icol=icol+1
          fstat = max(fstat,status)                     
          call  FTPCLD(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ra         ,status)
          icol=icol+1
          fstat = max(fstat,status)                         
          call  FTPCLD(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%dec        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                     
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%radec_err  ,status)
          icol=icol+1
          fstat = max(fstat,status)                    
          call  FTPCLD(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%lII        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                       
          call  FTPCLD(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%bII        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                       
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%rawx       ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%rawy       ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%offax      ,status) 
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ccdnr      ,status)
          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr1        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                       
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr1_err    ,status) 
          icol=icol+1
          fstat = max(fstat,status)                     
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr2        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                       
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr2_ERR    ,status) 
          icol=icol+1
          fstat = max(fstat,status)                    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr3        ,status) 
          icol=icol+1
          fstat = max(fstat,status)                       
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%hr3_err    ,status) 
          icol=icol+1
          fstat = max(fstat,status)                     
          ! The image / photon extraction radius is stored in %ecutrad (15/03/2018)
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ecutrad     ,status) 
          icol=icol+1
          fstat = max(fstat,status)     
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%maskfrac   ,status) 
          icol=icol+1
          fstat = max(fstat,status)    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%eef        ,status) 
          icol=icol+1
          fstat = max(fstat,status)    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%vignetting ,status) 
          icol=icol+1
          fstat = max(fstat,status)    
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ontime     ,status) 
          icol=icol+1
          fstat = max(fstat,status) 
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%dist_nn     ,status) 
          icol=icol+1
          fstat = max(fstat,status) 
!          call  FTPCLS(unit,45,i1,1,rowbuf,mllist%data(i1:i2)%flag         ,status) 
!          fstat = max(fstat,status)  
          call  FTPCLB(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%owner       ,status)
          icol=icol+1
          fstat = max(fstat,status)
          !*****************************************
          ! AGE HAS ADDED APETOOL RELATED PARAMETERS
          call  FTPCLJ(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_cts      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_bkg      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_exp      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_eef      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_radius      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          call  FTPCLE(unit,icol,i1,1,rowbuf,mllist%data(i1:i2)%ape_pois      ,status)
          fstat = max(fstat,status)  
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif ! 

  end subroutine fput_mllist

!-------------------------------------------------------------------------------

  subroutine fget_boxlist(filename,boxlist,ierr)

    implicit none
  
    type(boxlist_type)               :: boxlist
    character(len=*), intent(in)     :: filename
    character(len=7)                 :: extname='SRCLIST'
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(boxlist%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_BOXLIST'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
     
     
  !test:
 !  write(*,*) rtname, ' 1) ',boxlist%cname(1), boxlist%cname(33), boxlist%ncol


    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTNOPN(unit,filename,rwmode,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      else  
!        allocate(boxlist) 
        call fread_stdhdr(unit,boxlist%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,extname,0,status)
        if (status /= 0) then
          ierr = 3
          errstr = 'error moving to extension '//extname 
          call WRERR(rtname,ierr,errstr)
        else 
    
          call fread_shapehdr(unit,boxlist%shapehdr,status)
          nrows=boxlist%shapehdr%naxis2
          allocate(boxlist%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          do icol=1,boxlist%ncol
            call FTGCNO(unit,.false.,boxlist%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//boxlist%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo

          do while (nrows > irow)  
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            icol=1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,boxlist%data(i1:i2)%id_src ,fnull,status)
            ! copy id_src values to box_id_src:
            boxlist%data(i1:i2)%box_id_src=boxlist%data(i1:i2)%id_src
            fstat = max(fstat,status)
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,boxlist%data(i1:i2)%id_inst    ,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,boxlist%data(i1:i2)%id_band    ,fnull,status)
            fstat = max(fstat,status)             
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%scts       ,fnull,status)
            fstat = max(fstat,status)                                 
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%scts_err   ,fnull,status)
            fstat = max(fstat,status)                         
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%box_cts    ,fnull,status)
            fstat = max(fstat,status)                         
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%x_ima      ,fnull,status) 
            fstat = max(fstat,status)                            
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%x_ima_err  ,fnull,status)
            fstat = max(fstat,status)                      
            icol=icol+1
            if (colnum(icol) > 0)  call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%y_ima      ,fnull,status) 
            fstat = max(fstat,status)                            
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%y_ima_err  ,fnull,status)
            fstat = max(fstat,status)                      
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%like       ,fnull,status)
            fstat = max(fstat,status)                               
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%bg_map     ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%bg_raw     ,fnull,status)
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%exp_map    ,fnull,status)
            fstat = max(fstat,status)                          
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%flux       ,fnull,status)
            fstat = max(fstat,status)                             
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%flux_err   ,fnull,status)
            fstat = max(fstat,status)                        
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%rate       ,fnull,status) 
            fstat = max(fstat,status)                          
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%rate_err   ,fnull,status)
            fstat = max(fstat,status)                       
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,boxlist%data(i1:i2)%ra         ,fnull,status)
            fstat = max(fstat,status)                             
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,boxlist%data(i1:i2)%dec        ,fnull,status) 
            fstat = max(fstat,status)                         
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%radec_err  ,fnull,status)
            fstat = max(fstat,status)                      
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,boxlist%data(i1:i2)%lII        ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,boxlist%data(i1:i2)%bII        ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr1        ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr1_err    ,fnull,status) 
            fstat = max(fstat,status)                       
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr2        ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr2_err    ,fnull,status) 
            fstat = max(fstat,status)                      
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr3        ,fnull,status) 
            fstat = max(fstat,status)                           
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%hr3_err    ,fnull,status) 
            fstat = max(fstat,status)                       
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%vignetting ,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,boxlist%data(i1:i2)%box_size   ,fnull,status) 
            fstat = max(fstat,status)    
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%eef        ,fnull,status) 
            fstat = max(fstat,status)                         
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%dist_nn    ,fnull,status) 
            fstat = max(fstat,status) 
            ! in case column ext is present,it is also read: (GL 12/10/2018)
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,boxlist%data(i1:i2)%ext    ,fnull,status) 
            boxlist%has_extentcol=.false.
            if (status == 0) boxlist%has_extentcol=.true.
            fstat = max(fstat,status) 


            if (fstat /= 0) then
              ierr = 2
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif

          end do
        endif 
      endif
      status=0
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif

  end subroutine fget_boxlist

!-------------------------------------------------------------------------------

  subroutine fput_boxlist(filename,boxlist,ierr)

    implicit none
 
  
    type(boxlist_type)                 :: boxlist
    character(len=*), intent(in)       :: filename
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: nhdu, varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=33
    character(len=16), dimension(ncol) :: ttype,tunit
    character(len=4),  dimension(ncol) :: tform
    character(len=16)                  :: extname
    character(len=*),parameter         :: rtname = 'FPUT_BOXLIST'
    character(len=80)                  :: errstr

    extname='SRCLIST'

   ! NOTE: All elements of character array need to be assigned by strings of equal length: 

    ttype=[ character(len=16) :: 'id_src','id_inst','id_band','scts','scts_err','box_cts','x_ima','x_ima_err',  &
            'y_ima','y_ima_err','like','bg_map','bg_raw','exp_map','flux','flux_err',  &
            'rate','rate_err','ra','dec','radec_err','lII','bII',  &
            'hr1','hr1_err','hr2','hr2_er','hr3','hr3_err','vignetting',  &
            'box_size','eef','dist_nn']

    tform=[ character(len=4) :: '1J  ','1J  ','1J  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', &
            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ', & 
            '1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1E  ','1J  ','1E  ','1E  ']

    
    tunit=[ character(len=16) :: '','','','','','','','','','','','','', &
            '','','','','','','','','','',          &
            '','','','','','','','','','']
    

    status = 0
    fstat  = 0
    ierr   = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=boxlist%shapehdr%naxis2
    
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    call FTINIT(unit,filename,blocksize,status)
    fstat=max(fstat,status)
    if (fstat /= 0) then    
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else    
      status   = 0
      simple   = .true.
      bitpix   = 16
      naxis    = 0
      naxes(1) = 0
      !write FITS standard header:   
      call FTPHPS(unit,bitpix,naxis,naxes,status) 
      fstat=max(fstat,status)  
      ! write eROSITA standard header to primary extension (obsolete?? 29/04/2015):
      call fwrite_stdhdr(unit,boxlist%stdhdr,status)
     status = 0
      ! create new empty HDU:
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  

      ! write header for binary table:
      call FTPHBN(unit,nrows,ncol,ttype,tform,tunit,extname,varidat,status)
      fstat=max(fstat,status)  

      ! write eROSITA standard header to table extension (29/04/2015):
      call fwrite_stdhdr(unit,boxlist%stdhdr,status)
      fstat=max(fstat,status)  

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
     
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 

        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2        
          call FTPCLJ(unit,1,i1,1,rowbuf,boxlist%data(i1:i2)%id_src, status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,2,i1,1,rowbuf,boxlist%data(i1:i2)%id_inst    ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,3,i1,1,rowbuf,boxlist%data(i1:i2)%id_band    ,status)
          fstat = max(fstat,status)               
          call  FTPCLE(unit,4,i1,1,rowbuf,boxlist%data(i1:i2)%scts       ,status)
          fstat = max(fstat,status)                                 
          call  FTPCLE(unit,5,i1,1,rowbuf,boxlist%data(i1:i2)%scts_err   ,status)
          fstat = max(fstat,status)                         
          call  FTPCLE(unit,6,i1,1,rowbuf,boxlist%data(i1:i2)%box_cts    ,status)
          fstat = max(fstat,status)                         
          call  FTPCLE(unit,7,i1,1,rowbuf,boxlist%data(i1:i2)%x_ima      ,status) 
          fstat = max(fstat,status)                            
          call  FTPCLE(unit,8,i1,1,rowbuf,boxlist%data(i1:i2)%x_ima_err  ,status)
          fstat = max(fstat,status)                        
          call  FTPCLE(unit,9,i1,1,rowbuf,boxlist%data(i1:i2)%y_ima      ,status) 
          fstat = max(fstat,status)                              
          call  FTPCLE(unit,10,i1,1,rowbuf,boxlist%data(i1:i2)%y_ima_err  ,status)
          fstat = max(fstat,status)                      
          call  FTPCLE(unit,11,i1,1,rowbuf,boxlist%data(i1:i2)%like       ,status)
          fstat = max(fstat,status)                               
          call  FTPCLE(unit,12,i1,1,rowbuf,boxlist%data(i1:i2)%bg_map     ,status) 
          fstat = max(fstat,status)                           
          call  FTPCLE(unit,13,i1,1,rowbuf,boxlist%data(i1:i2)%bg_raw     ,status)
          fstat = max(fstat,status)                           
          call  FTPCLE(unit,14,i1,1,rowbuf,boxlist%data(i1:i2)%exp_map    ,status)
          fstat = max(fstat,status)                          
          call  FTPCLE(unit,15,i1,1,rowbuf,boxlist%data(i1:i2)%flux       ,status)
          fstat = max(fstat,status)                             
          call  FTPCLE(unit,16,i1,1,rowbuf,boxlist%data(i1:i2)%flux_err   ,status)
          fstat = max(fstat,status)                        
          call  FTPCLE(unit,17,i1,1,rowbuf,boxlist%data(i1:i2)%rate       ,status) 
          fstat = max(fstat,status)                          
          call  FTPCLE(unit,18,i1,1,rowbuf,boxlist%data(i1:i2)%rate_err   ,status)
          fstat = max(fstat,status)                       
          call  FTPCLD(unit,19,i1,1,rowbuf,boxlist%data(i1:i2)%ra         ,status)
          fstat = max(fstat,status)                             
          call  FTPCLD(unit,20,i1,1,rowbuf,boxlist%data(i1:i2)%dec        ,status) 
          fstat = max(fstat,status)                         
          call  FTPCLE(unit,21,i1,1,rowbuf,boxlist%data(i1:i2)%radec_err  ,status)
          fstat = max(fstat,status)                      
          call  FTPCLD(unit,22,i1,1,rowbuf,boxlist%data(i1:i2)%lII        ,status) 
          fstat = max(fstat,status)                           
          call  FTPCLD(unit,23,i1,1,rowbuf,boxlist%data(i1:i2)%bII        ,status) 
            fstat = max(fstat,status)                           
          call  FTPCLE(unit,24,i1,1,rowbuf,boxlist%data(i1:i2)%hr1        ,status) 
          fstat = max(fstat,status)                           
          call  FTPCLE(unit,25,i1,1,rowbuf,boxlist%data(i1:i2)%hr1_err    ,status) 
          fstat = max(fstat,status)                       
          call  FTPCLE(unit,26,i1,1,rowbuf,boxlist%data(i1:i2)%hr2        ,status) 
          fstat = max(fstat,status)                           
          call  FTPCLE(unit,27,i1,1,rowbuf,boxlist%data(i1:i2)%hr2_ERR    ,status) 
          fstat = max(fstat,status)                      
          call  FTPCLE(unit,28,i1,1,rowbuf,boxlist%data(i1:i2)%hr3        ,status) 
          fstat = max(fstat,status)                           
          call  FTPCLE(unit,29,i1,1,rowbuf,boxlist%data(i1:i2)%hr3_err    ,status) 
          fstat = max(fstat,status)                       
          call  FTPCLE(unit,30,i1,1,rowbuf,boxlist%data(i1:i2)%vignetting ,status) 
          fstat = max(fstat,status)                   
          call  FTPCLJ(unit,31,i1,1,rowbuf,boxlist%data(i1:i2)%box_size   ,status) 
          fstat = max(fstat,status)                      
          call  FTPCLE(unit,32,i1,1,rowbuf,boxlist%data(i1:i2)%eef        ,status) 
          fstat = max(fstat,status)                         
          call  FTPCLE(unit,33,i1,1,rowbuf,boxlist%data(i1:i2)%dist_nn    ,status) 
          fstat = max(fstat,status) 
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif ! 

  end subroutine fput_boxlist


!-------------------------------------------------------------------------------

  subroutine sort_files(headers,nimage,instrume,first,detmaskflag,ierr)

  !---------------------------------------------------------------------
  ! Provides part of the functionality of XMM-SAS routine 'get_files'
  !  !!!!!!!!!!!!!!!!!!
  ! NOW OBSOLETE (use sort_ima to create IP and EXPIP)
  !  !!!!!!!!!!!!!!!!!!
  !---------------------------------------------------------------------
 
    integer(kind=int32)                      :: nimage,ierr  
    character(len=clong),dimension(max_ima)  :: files
    character(len=21),dimension(:)           :: instrume
    logical(kind=log32)                      :: first,detmaskflag
    integer(kind=int32)                      :: i, j, ib, status
    integer (kind=int32)                     :: ipoint 
    logical(kind=log32), dimension(max_ima)  :: double
    character(len=21)  , dimension(max_ima)  :: ins
    type(stdhdr_type)                        :: header
    character(len=*), parameter              :: rtname='SORT_FILES'
    ierr=0
    status=0

    if (first) then

      do i=1,nfiles
        call fget_stdhdr(files(i),header,status)
        instrume(i) = trim(header%telescop) // trim(header%instrume) // trim(header%exp_id)
      enddo

      num_instr = 0
      double = .false.    
      ip = -1
      do i=1,nimage        
        if (.not. double(i)) num_instr = num_instr+1
        ib = 1 
        do while (ip(num_instr,ib) /= -1)
          ib = ib+1
        end do
        ip(num_instr,ib) = i
        num_image(num_instr) = ib
        do j=i+1,nimage
          if (instrume(i) == instrume(j)) double(j) = .true.
        end do
      end do

      do i=1,num_instr
        ipoint=ip(i,1)
        expip(i)=-1
        do j=1,max_cam
          if (index(instrume(ipoint),cam_names(j)) > 0 ) then
            expip(i) = j
          endif
        end do
        if (expip(i) == -1) then
          ierr = 2
        end if
      end do
        
    else !(first)

      if (detmaskflag) then
        if (nimage /= num_instr) then
          ierr = 2
        end if
      else 
        if (nimage /= nim) then
           ierr = 2
        end if
      end if
 
      do i=1,nfiles
        call fget_stdhdr(files(i),header,status)
        ins(i) = trim(header%telescop) // trim(header%instrume) // trim(header%exp_id)
      enddo
      
      if (detmaskflag) then 
         do i=1,num_instr
           do j=1,num_image(i)
             ipoint=ip(i,j)
             if (ins(i) /= instrume(ipoint)) then 
               ierr = 2
             end if
           end do
         end do
      end if 

    endif 
    
    nim = ip(num_instr,num_image(num_instr))


  !OBSOLETE
  end subroutine sort_files
  


!-------------------------------------------------------------------------------

  subroutine sort_ima(headers,mask_headers,nimage,nmask,ierr)

  !------------------------------------------------------------
  ! Create index arrays IP and EXPIP from array of
  ! standard eROSITA science image headers
  !
  ! Use of global parameters:
  !   MAX_IMA
  !
  ! Set global variables:
  !   IP
  !   EXPIP
  !   BANDIP
  !   INSTIP
  !   NUM_EXP
  !   NUM_IMAGE
  !------------------------------------------------------------

    type(stdhdr_type),dimension(:),intent(in) :: headers
    type(stdhdr_type),dimension(:),intent(in) :: mask_headers
    integer(kind=int32),intent(in)            :: nimage, nmask  
    integer(kind=int32),intent(out)           :: ierr 
    integer(kind=int32)                       :: i, j, iexp
    integer(kind=int32),dimension(max_exp)    :: nbands
    integer (kind=int32)                      :: ipoint,icam,ncam,found
    integer(kind=int32), dimension(max_ima)   :: double
    character(len=*), parameter               :: rtname='SORT_IMA'
 
    ierr=0

    num_exp = 0 
    double = 0    
    ip(:,:) = -1
    mask_ip(:) = -1
    num_image(:) = 0
    bandip(:) = 0
    expip(:)=0
    instip(:)=0

!    write(*,*) 'headers(1)%instruments : ',headers(1)%instruments

  ! find all instruments (or combinations) present:      
    icam=0
    ncam=0
    cam_names(1)=headers(1)%instruments
    ncam=1
    do i=2,nimage
      found=0
      do icam=1,ncam
        if (headers(i)%instruments == cam_names(icam)) found=found+1
      enddo
      if (found == 0) then 
        ncam=ncam+1
        cam_names(ncam) = headers(i)%instruments
      endif
    enddo



    do i=1,nimage 
      if (expip(i) == 0) then 
        ! found new exposure
        num_exp = num_exp+1
        iexp = num_exp
        expip(i) = iexp
        bandip(i) = 1
        num_image(iexp) = 1

      else
        ! new energy band in existing exposure
        iexp=expip(i)
        num_image(iexp)=num_image(iexp)+1
        bandip(i)=num_image(iexp)
      endif
     
      ip(expip(i),bandip(i)) = i

      do j=i+1,nimage
        if (match_exposures(headers(i),headers(j)) .and. expip(j) == 0) then
          
          expip(j) = expip(i)
        endif
      end do
      do j=1,nmask
        
        if (match_exposures(headers(i),mask_headers(j)))  mask_ip(i) = j
!        write(*,*) 'match headers:',i,j, mask_ip(i)
      enddo
    end do

    do i=1,nimage
      if (mask_ip(i) == -1 ) then
        ierr=3
!        write(*,*) 'nimage:',nimage
        call wrerr(rtname,ierr,'No detection mask for exposure' )
      endif
    enddo
   
    do i=1,num_exp
      ipoint=ip(i,1)
      instip(i)=-1

      do j=1,num_exp
        if (headers(ipoint)%instruments == cam_names(j)) then
          instip(i) = j
        endif
      end do
      if (instip(i) == -1) then
        ierr = 2
      end if
    end do
        


  end subroutine sort_ima


!--------------------------------------------------------------------------------


  subroutine filter_srclist(par,nsrc,srclist,ierr)

  !--------------------------------------------------------------------
  ! Filter erbox srclist for likelihood threshold 
  !
  !--------------------------------------------------------------------
   

    type(erbox_param_type)                     :: par
    integer(kind=int32)                        :: nsrc
    type(srcrow_type),dimension(:,:)           :: srclist
    integer(kind=int32),intent(out)            :: ierr
    integer(kind=int32)                        :: status
    integer(kind=int32)                        :: iima,isrc,id_src
    real(kind=real32)                          :: likesum,liketot,ndof

    character(len=*),parameter         :: rtname = 'FILTER_SRCLIST'

    ierr = 0 
    status = 0
    
    id_src = 0
 

    do isrc = 1,nsrc
      ndof    = 0.0
      likesum = 0.0
      do iima = 1,par%n_imagesets
        if (srclist(iima,isrc)%maskzero  == 0) then
          likesum = likesum + srclist(iima,isrc)%like
          ndof = ndof + 1.0 
        endif
      enddo
      liketot =  gammq(ndof, likesum)
      if (liketot >= par%likemin) then
        id_src = id_src + 1
        srclist(:,id_src) = srclist(:,isrc)
      endif
    end do
    
    nsrc = id_src


  end subroutine filter_srclist

!--------------------------------------------------------------------------------


  subroutine format_boxlist(par,image,nsrc,srclist,boxlist,ierr)

  !--------------------------------------------------------------
  ! Re-format erbox source list from 2-dimensional form
  ! (nim,nsrc) to XMM style output list format: 
  ! number of rows per source:
  ! n_rows=(n_bands+sum_bands)*n_cameras+sum_cameras 
  ! Uses global variables: NUM_EXP, NIM, MAX_CAM, IP
  ! NOTE: Do not give attribute intent(out) to structure containing 
  !       pointer (e.g. boxlist) 
  !--------------------------------------------------------------

    implicit none

    type(erbox_param_type)                     :: par
    type(i4cube_type)                          :: image
    integer(kind=int32)                        :: nsrc
    type(srcrow_type),dimension(:,:), pointer  :: srclist
    type(boxlist_type)                         :: boxlist
    integer(kind=int32),intent(out)            :: ierr
    integer(kind=int32)                        :: status
    integer(kind=int32)                        :: nzero   
    integer(kind=int32)                        :: iima,isrc,iexp,iinst,ipoint
    type(srcrow_type),dimension(:,:,:), pointer:: srcbuff 
    type(srcrow_type), pointer                 :: plist,pbuff,pband,pinst,pbox
    type(srcrow_type),dimension(:,:), pointer  :: bandsum
    type(srcrow_type),dimension(:),   pointer  :: instsum
    logical(kind=log32),dimension(max_cam)     :: instflag
    integer(kind=int32),dimension(max_cam)     :: nima_inst 
    integer(kind=int32)                        :: id_src,irow                  

    character(len=*),parameter                 :: rtname = 'FORMAT_BOXLIST'

    ierr=0
    status=0
   
 !   allocate(srcbuff(max_cam,max_ima,nsrc))



    allocate(srcbuff(max_cam,nim,nsrc))

    call init_boxlist3d(srcbuff)

    allocate(instsum(nsrc))
    call init_boxlist1d(instsum)

    allocate(bandsum(max_cam,nsrc))
    call init_boxlist2d(bandsum)  

    instflag(1:max_cam)=.false.
    nima_inst(:)=0
    id_src = 0
    irow=0


    do isrc=1,nsrc
      pinst => instsum(isrc) 
      do iexp=1,num_exp
        iinst=instip(iexp)
        instflag(iinst)=.true.
        do iima=1,num_image(iexp)
          ipoint = ip(iexp,iima)
          plist => srclist(ipoint,isrc)
          pbuff => srcbuff(iinst,iima,isrc)
          pband => bandsum(iinst,isrc)
          
                    

          if (plist%maskzero == 0) then
            ! calculate rates and fluxes:
            if (plist%exp_map > 0.0)  then
              plist%rate     = plist%scts     / plist%exp_map
              plist%rate_err = plist%scts_err / plist%exp_map
            else
              plist%rate     = 0.0 
              plist%rate_err = 0.0  
            endif       
            if (par%ecf(ipoint) > 0.0)  then
              plist%flux     = plist%rate     / par%ecf(ipoint)
              plist%flux_err = plist%rate_err / par%ecf(ipoint)
            else
              plist%flux =      0.0
              plist%flux_err =  0.0
! needs gfortran version >= 5:    
!              plist%flux =     ieee_value(plist%flux,ieee_quite_nan)
!              plist%flux_err = ieee_value(plist%flux_err,ieee_quite_nan)
             endif
 
            
 
            ! source parameters - single energy bands:           
            call addwitherr(pbuff%scts,pbuff%scts_err,plist%scts,plist%scts_err,pbuff%scts,pbuff%scts_err,status)
            pbuff%id_inst  =   iinst
            pbuff%id_band  =   bandip(ipoint)
            pbuff%box_cts  =   pbuff%box_cts  + plist%box_cts
            pbuff%bg_raw   =   pbuff%bg_raw   + plist%bg_raw
            pbuff%eef      =   pbuff%eef      + plist%eef  
            pbuff%like     =   pbuff%like     + plist%like  
            pbuff%ndof     =   pbuff%ndof     + 1
            pbuff%exp_map  =   pbuff%exp_map  + plist%exp_map
            pbuff%bg_map  =   pbuff%bg_map    + plist%bg_map                   
            ! exposure weighted averages of  fluxes:
            pbuff%flux     =   pbuff%flux     + plist%flux*plist%exp_map
            pbuff%flux_err =   pbuff%flux_err + (plist%flux_err*plist%exp_map)**2
            !  summary of energy bands:
            call addwitherr(pband%scts,pband%scts_err, plist%scts,plist%scts_err,pband%scts,pband%scts_err,status)  
            pband%id_inst  =   iinst  
            pband%id_band  =   0              
            pband%rate     =   pband%rate     + plist%rate*plist%exp_map
            pband%rate_err =   pband%rate_err + (plist%rate_err*plist%exp_map)**2   
            pband%box_cts  =   pband%box_cts  + plist%box_cts
            pband%bg_raw   =   pband%bg_raw   + plist%bg_raw  
            pband%bg_map   =   pband%bg_map   + plist%bg_map 
!            pband%exp_map  =   pband%exp_map  + plist%exp_map 
            pband%exp_map  =   0.0 
            pband%like     =   pband%like     + plist%like  
            pband%ndof     =   pband%ndof + 1                                                   
            ! summary of instruments:
            call addwitherr(pinst%scts,pinst%scts_err, plist%scts,plist%scts_err,pinst%scts,pinst%scts_err,status)
            pinst%id_inst  =   0
            pinst%id_band  =   0            
            pinst%box_cts  =   pinst%box_cts  + plist%box_cts
            pinst%bg_raw   =   pinst%bg_raw   + plist%bg_raw  
            pinst%bg_map   =   pinst%bg_map   + plist%bg_map 
!            pinst%exp_map  =   pinst%exp_map  + plist%exp_map 
            pinst%exp_map  =   0.0 
            pinst%like     =   pinst%like     + plist%like
            pinst%ndof     =   pinst%ndof     + 1
            ! count maximum number of bands per instrument:
            nima_inst(iinst)=max(iima,nima_inst(iinst))
          endif

 
           nullify(plist)
           nullify(pbuff)
           nullify(pband)
         enddo ! num_image
      enddo !num_exp
 
      ! averages over exposures for rates and fluxes:
      do iinst=1,max_cam
        pband => bandsum(iinst,isrc)
         if (instflag(iinst)) then
          do iima=1,nima_inst(iinst)
            pbuff => srcbuff(iinst,iima,isrc)
!            pbuff%scts_err = SQRT(pbuff%scts_err)
            pbuff%x_ima = srclist(1,isrc)%x_ima
            pbuff%x_ima_err = srclist(1,isrc)%x_ima_err
            pbuff%y_ima = srclist(1,isrc)%y_ima
            pbuff%y_ima_err = srclist(1,isrc)%y_ima_err
            pbuff%box_size = srclist(1,isrc)%box_size
            if (pbuff%exp_map > 0.0) then
              pbuff%rate     = pbuff%scts/pbuff%exp_map
              pbuff%rate_err = pbuff%scts_err/pbuff%exp_map
              pbuff%flux     = pbuff%flux/pbuff%exp_map
              pbuff%flux_err = SQRT(pbuff%flux_err)/pbuff%exp_map 
              pbuff%like     = gammq(real(pbuff%ndof), pbuff%like)
              pband%rate     = pband%rate +   pbuff%rate 
              pband%rate_err = pband%rate_err +   (pbuff%rate_err)**2 
              pband%flux     = pband%flux +   pbuff%flux 
              pband%flux_err = pband%flux_err +   (pbuff%flux_err)**2  
            endif
            nullify(pbuff)
          enddo
          pband%rate_err = SQRT(pband%rate_err) 
          pband%flux_err = SQRT(pband%flux_err) 
          pband%like     = gammq(real(pband%ndof), pband%like)  
          pband%x_ima = srclist(1,isrc)%x_ima
          pband%x_ima_err = srclist(1,isrc)%x_ima_err
          pband%y_ima = srclist(1,isrc)%y_ima
          pband%y_ima_err = srclist(1,isrc)%y_ima_err
          pband%box_size = srclist(1,isrc)%box_size         

          if (par%hrdef(1)>0.and.par%hrdef(2)>0 .and.  max(par%hrdef(1),par%hrdef(2)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(1),isrc)%rate,srcbuff(iinst,par%hrdef(1),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(2),isrc)%rate,srcbuff(iinst,par%hrdef(2),isrc)%rate_err, &
                        pband%hr1,pband%hr1_err,status)
          endif              
          if (par%hrdef(3)>0.and.par%hrdef(4)>0 .and.  max(par%hrdef(3),par%hrdef(3)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(3),isrc)%rate,srcbuff(iinst,par%hrdef(4),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(3),isrc)%rate,srcbuff(iinst,par%hrdef(4),isrc)%rate_err, &
                        pband%hr2,pband%hr2_err,status)
          endif 
          if (par%hrdef(5)>0.and.par%hrdef(6)>0 .and.  max(par%hrdef(5),par%hrdef(6)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(5),isrc)%rate,srcbuff(iinst,par%hrdef(6),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(5),isrc)%rate,srcbuff(iinst,par%hrdef(6),isrc)%rate_err, &
                        pband%hr3,pband%hr3_err,status)
          endif 
          pinst%rate     = pinst%rate + pband%rate 
          pinst%rate_err = pinst%rate_err + (pband%rate_err)**2
          pinst%flux     = pinst%flux + pband%flux/(pband%flux_err)**2 
          pinst%flux_err = pinst%flux_err + (pband%flux_err)**2
         endif

        nullify(pband)

      enddo
      pinst%rate_err  = SQRT(pinst%rate_err)     
      pinst%flux      = pinst%flux/pinst%flux_err
      pinst%flux_err  = 1./SQRT(pinst%flux_err)  
      pinst%like      = gammq(real(pinst%ndof), pinst%like)   
      pinst%x_ima = srclist(1,isrc)%x_ima
      pinst%x_ima_err = srclist(1,isrc)%x_ima_err
      pinst%y_ima = srclist(1,isrc)%y_ima
      pinst%y_ima_err = srclist(1,isrc)%y_ima_err
      pinst%box_size = srclist(1,isrc)%box_size         

      ! copy to erbox output list:
      if (pinst%like >= par%likemin) then 
        id_src =  id_src + 1
        irow = irow + 1
!        boxlist%data(irow)%box_id_src = id_src 
        pinst%id_src = id_src
        boxlist%data(irow) = pinst
        do iinst=1,max_cam
          if (instflag(iinst)) then
            irow=irow+1
            boxlist%data(irow) =  bandsum(iinst,isrc)
            boxlist%data(irow)%id_src = id_src 
            do iima=1,nima_inst(iinst)
              irow=irow+1
              boxlist%data(irow) = srcbuff(iinst,iima,isrc)
              boxlist%data(irow)%id_src = id_src
            enddo
          endif
        enddo
      endif
      
      nullify(pinst)
    enddo !nsrc
    
    boxlist%shapehdr%naxis2 = irow 

    
    ! calculate celestial coordinates:

    do irow=1, boxlist%shapehdr%naxis2
      plist =>  boxlist%data(irow) 
      call  pix2equ(image%wcshdr(1), plist%x_ima, plist%x_ima_err ,plist%y_ima, plist%y_ima_err, &
                    plist%ra, plist%dec, plist%radec_err,ierr)
    enddo    

    nullify(plist)


    call fput_boxlist(par%boxlistset,boxlist,status)
    


  end subroutine format_boxlist
   
 !--------------------------------------------------------------------------------
  

  subroutine format_mllist(par,image,nsrc,srclist,mllist,ierr)

  !--------------------------------------------------------------
  ! Re-format erml source list from 2-dimensional form (srclist)
  ! (nim,nsrc) to XMM style output list format (mllist): 
  ! number of rows per source:
  ! n_rows=(n_bands+sum_bands)*n_cameras+sum_cameras 
  ! NOTE: Do not give attribute intent(out) to structure containing 
  !       pointer (e.g. mllist) 
  !--------------------------------------------------------------

    implicit none

    type(ermldet_param_type)                   :: par
    type(i4cube_type)                          :: image
    integer(kind=int32)                        :: nsrc
    type(srcrow_type),dimension(:,:),pointer   :: srclist
    type(mllist_type)                          :: mllist
    integer(kind=int32),intent(out)            :: ierr
    integer(kind=int32)                        :: status
    integer(kind=int32)                        :: nzero   
    integer(kind=int32)                        :: iima,isrc,iexp,iinst,ipoint
    type(srcrow_type),dimension(:,:,:),pointer :: srcbuff 
    type(srcrow_type), pointer                 :: plist,pbuff,pband,pinst,pbox
    type(srcrow_type),dimension(:,:), pointer  :: bandsum
    type(srcrow_type),dimension(:),   pointer  :: instsum
    logical(kind=log32),dimension(max_cam)     :: instflag
    integer(kind=int32),dimension(max_cam)     :: nima_inst 
    integer(kind=int32)                        :: id_src,irow,isumm ,ndof_src ,owner              
    real(kind=real32)                          :: pix2arcsec,pix2arcmin

    character(len=*),parameter                 :: rtname = 'FORMAT_MLLIST'

    ierr=0
    status=0
   

    allocate(srcbuff(max_cam,nim,nsrc))
    call init_mllist3d(srcbuff)

    allocate(instsum(nsrc))
    call init_mllist1d(instsum)

    allocate(bandsum(max_cam,nsrc))
    call init_mllist2d(bandsum)  

    
    pix2arcsec=(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2))*1800.0
    pix2arcmin=pix2arcsec/60.0

    srclist(:,:)%ecutrad=srclist(:,:)%ecutrad*pix2arcsec
    srclist(:,:)%ext=srclist(:,:)%ext*pix2arcsec
    srclist(:,:)%ext_err=srclist(:,:)%ext_err*pix2arcsec

    srclist(:,:)%bg_map=srclist(:,:)%bg_map/pix2arcmin/pix2arcmin


    ! summary over images element in srclist array (0 or 1)
    isumm = 0

    instflag(1:max_cam)=.false.
    nima_inst(:)=0
    id_src = 0
    irow=0


    do isrc=1,nsrc
      ndof_src=srclist(0,isrc)%ndof
      pinst => instsum(isrc)
      pinst%ndof =  srclist(0,isrc)%ndof
      do iexp=1,num_exp
        iinst=instip(iexp)
        instflag(iinst)=.true.
        pband => bandsum(iinst,isrc)
        pband%ndof = srclist(0,isrc)%ndof
        do iima=1,num_image(iexp)
          ipoint = ip(iexp,iima)
          plist => srclist(ipoint,isrc)
          pbuff => srcbuff(iinst,iima,isrc)
 
          if (plist%maskzero == 0) then 
            plist%scts     = plist%rate     * plist%exp_map
            plist%scts_err = plist%rate_err * plist%exp_map
            if (par%ecf(ipoint) > 0.0)  then
              plist%flux     = plist%rate     / par%ecf(ipoint)
              plist%flux_err = plist%rate_err / par%ecf(ipoint)
            else 

              plist%flux     = 0.0
              plist%flux_err = 0.0
!        nedds gfortran version >= 5:
!              plist%flux     = ieee_value(plist%flux,ieee_quite_nan)
!              plist%flux_err = ieee_value(plist%flux_err,ieee_quite_nan)
            endif
          
            ! source parameters - single energy bands:    

            call addwitherr(pbuff%scts,pbuff%scts_err,plist%scts,plist%scts_err,pbuff%scts,pbuff%scts_err,status)
            pbuff%id_inst  =   iinst
            pbuff%id_band  =   bandip(ipoint)
            pbuff%eef      =   plist%eef  
            pbuff%ecutrad   =   plist%ecutrad
            pbuff%like     =   pbuff%like     + plist%like  
            pbuff%ndof     =   pbuff%ndof     + 1
            pbuff%exp_map  =   pbuff%exp_map  + plist%exp_map
            pbuff%bg_map  =    pbuff%bg_map   + plist%bg_map                   
            ! exposure weighted averages of fluxes:
            pbuff%flux     =   pbuff%flux     + plist%flux*plist%exp_map
            pbuff%flux_err =   pbuff%flux_err + (plist%flux_err*plist%exp_map)**2
            !  summary of energy band00s:

            call addwitherr(pband%scts,pband%scts_err, plist%scts,plist%scts_err,pband%scts,pband%scts_err,status)  
            pband%id_inst  =   iinst  
            pband%id_band  =   0   


!            pband%rate     =   pband%rate     + plist%rate
!            pband%rate_err =   pband%rate_err + (plist%rate_err)**2

!!!!        This should not be added, (?) (GL 02-08-2012):   
            pband%bg_map   =   pband%bg_map   + plist%bg_map 
!!!1
!            pband%exp_map  =   pband%exp_map  + plist%exp_map
            pband%exp_map  =   0.0
  
            pband%like     =   pband%like     + plist%like  
            pband%ndof     =   pband%ndof + 1 
            pband%ecutrad  =   max(pband%ecutrad,plist%ecutrad)
                                                
            ! summary of instruments:

            call addwitherr(pinst%scts,pinst%scts_err, plist%scts,plist%scts_err,pinst%scts,pinst%scts_err,status)
            pinst%id_inst  =   0
            pinst%id_band  =   0            
            pinst%bg_raw   =   pinst%bg_raw   + plist%bg_raw  
            pinst%bg_map   =   pinst%bg_map   + plist%bg_map 
!            pinst%exp_map  =   pinst%exp_map  + plist%exp_map
            pinst%exp_map  =   0.0
            pinst%like     =   pinst%like     + plist%like
            pinst%ecutrad  =   max(pinst%ecutrad,plist%ecutrad)
            pinst%ndof     =   pinst%ndof     + 1
      
            ! count maximum number of bands per instrument:
            nima_inst(iinst)=max(iima,nima_inst(iinst))
           endif

 
           nullify(plist)
           nullify(pbuff)
       !    nullify(pband)
         enddo ! num_image
         nullify(pband)
      enddo !num_exp


      ! averages over exposures for rates and fluxes:
      do iinst=1,max_cam
        pband => bandsum(iinst,isrc)
         if (instflag(iinst)) then
          do iima=1,nima_inst(iinst)
            pbuff => srcbuff(iinst,iima,isrc)
!           pbuff%scts_err = SQRT(pbuff%scts_err)
            pbuff%box_id_src = srclist(isumm,isrc)%box_id_src
            pbuff%x_ima      = srclist(isumm,isrc)%x_ima
            pbuff%x_ima_err  = srclist(isumm,isrc)%x_ima_err
            pbuff%y_ima      = srclist(isumm,isrc)%y_ima
            pbuff%y_ima_err  = srclist(isumm,isrc)%y_ima_err
            pbuff%ext        = srclist(isumm,isrc)%ext
            pbuff%ext_err    = srclist(isumm,isrc)%ext_err
            pbuff%ext_like   = srclist(isumm,isrc)%ext_like
            pbuff%id_cluster = srclist(isumm,isrc)%id_cluster  
            pbuff%ecutrad     = srclist(isumm,isrc)%ecutrad 


            if (pbuff%exp_map > 0.0) then
              ! exposure weighted average of rates from scts:
              pbuff%rate     = pbuff%scts/pbuff%exp_map
              pbuff%rate_err = pbuff%scts_err/pbuff%exp_map
              pbuff%flux     = pbuff%flux/pbuff%exp_map
              pbuff%flux_err = SQRT(pbuff%flux_err)/pbuff%exp_map 
!              write(*,*) rtname,' 00) ' , pbuff%like, pbuff%ndof+ndof_src
              ! divide by 2. => 2 degrees of freedom (GL 24/06/2014) 
              pbuff%like     = gammq(real(pbuff%ndof+ndof_src)/2., pbuff%like)
!              write(*,*) rtname,' 01 ' , pbuff%like, pbuff%ndof
              pband%rate     = pband%rate +   pbuff%rate 

              pband%rate_err = pband%rate_err +   (pbuff%rate_err)**2 
              pband%flux     = pband%flux +   pbuff%flux 
              pband%flux_err = pband%flux_err +   (pbuff%flux_err)**2  
            endif
            nullify(pbuff)
          enddo
          pband%rate_err = SQRT(pband%rate_err) 
          pband%flux_err = SQRT(pband%flux_err) 
!          write(*,*) rtname, ' 1) ',pband%like,pband%ndof 
          ! divide by 2. => 2 degrees of freedom (GL 24/06/2014)
          pband%like     = gammq(real(pband%ndof)/2., pband%like) 
!          write(*,*) rtname, ' 2) ',pband%like
          pband%box_id_src= srclist(isumm,isrc)%box_id_src
          pband%x_ima     = srclist(isumm,isrc)%x_ima
          pband%x_ima_err = srclist(isumm,isrc)%x_ima_err
          pband%y_ima     = srclist(isumm,isrc)%y_ima
          pband%y_ima_err = srclist(isumm,isrc)%y_ima_err
          pband%box_size  = srclist(isumm,isrc)%box_size    
          pband%ext       = srclist(isumm,isrc)%ext 
          pband%ext_err   = srclist(isumm,isrc)%ext_err 
          pband%ext_like  = srclist(isumm,isrc)%ext_like   
          pband%id_cluster  = srclist(isumm,isrc)%id_cluster   
    
          if (par%hrdef(1)>0.and.par%hrdef(2)>0 .and.  max(par%hrdef(1),par%hrdef(2)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(1),isrc)%rate,srcbuff(iinst,par%hrdef(1),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(2),isrc)%rate,srcbuff(iinst,par%hrdef(2),isrc)%rate_err, &
                        pband%hr1,pband%hr1_err,status)
          endif              
          if (par%hrdef(3)>0.and.par%hrdef(4)>0 .and.  max(par%hrdef(3),par%hrdef(3)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(3),isrc)%rate,srcbuff(iinst,par%hrdef(4),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(3),isrc)%rate,srcbuff(iinst,par%hrdef(4),isrc)%rate_err, &
                        pband%hr2,pband%hr2_err,status)
          endif 
          if (par%hrdef(5)>0.and.par%hrdef(6)>0 .and.  max(par%hrdef(5),par%hrdef(6)) <= nima_inst(iinst)) then
            call hrcalc(srcbuff(iinst,par%hrdef(5),isrc)%rate,srcbuff(iinst,par%hrdef(6),isrc)%rate_err, &
                        srcbuff(iinst,par%hrdef(5),isrc)%rate,srcbuff(iinst,par%hrdef(6),isrc)%rate_err, &
                        pband%hr3,pband%hr3_err,status)
          endif 
      
          pinst%rate     = pinst%rate + pband%rate 
          pinst%rate_err = pinst%rate_err + (pband%rate_err)**2
          pinst%flux     = pinst%flux + pband%flux/(pband%flux_err)**2 
          pinst%flux_err = pinst%flux_err + (1./pband%flux_err)**2
         endif
        nullify(pband)



      enddo
      pinst%rate_err  = SQRT(pinst%rate_err)     
      pinst%flux      = pinst%flux/pinst%flux_err
      pinst%flux_err  = 1./SQRT(pinst%flux_err)   
      pinst%like      = gammq(real(pinst%ndof)/2.0, pinst%like)
      pinst%box_id_src= srclist(isumm,isrc)%box_id_src 
      pinst%x_ima       = srclist(isumm,isrc)%x_ima
      pinst%x_ima_err   = srclist(isumm,isrc)%x_ima_err
      pinst%y_ima       = srclist(isumm,isrc)%y_ima
      pinst%y_ima_err   = srclist(isumm,isrc)%y_ima_err
      pinst%ext         = srclist(isumm,isrc)%ext 
      pinst%ext_err     = srclist(isumm,isrc)%ext_err 
      pinst%ext_like    = srclist(isumm,isrc)%ext_like   
      pinst%id_cluster  = srclist(isumm,isrc)%id_cluster   

      ! copy to ermldet output list:
      if (pinst%like >= par%likemin) then 
        id_src =  id_src + 1
        irow = irow + 1
!        mllist%data(irow)%box_id_src = id_src 
        pinst%id_src = id_src
        mllist%data(irow) = pinst
        do iinst=1,max_cam
          if (instflag(iinst)) then
            irow=irow+1
            mllist%data(irow) =  bandsum(iinst,isrc)
            mllist%data(irow)%id_src = id_src 
            do iima=1,nima_inst(iinst)
              irow=irow+1
              mllist%data(irow) = srcbuff(iinst,iima,isrc)
              mllist%data(irow)%id_src = id_src
            enddo
          endif
        enddo
      endif
      
      nullify(pinst)
    enddo !nsrc

    mllist%shapehdr%naxis2 = irow 

    ! caculate celestial coordinates:




    do irow=1, mllist%shapehdr%naxis2
      plist =>  mllist%data(irow) 
      call  pix2equ(image%wcshdr(1), plist%x_ima, plist%x_ima_err ,plist%y_ima, plist%y_ima_err, &
                    plist%ra, plist%dec, plist%radec_err,ierr)
      
      plist%owner=0
      call  calc_sourceowner(plist%ra,plist%dec,0.D0,plist%lII,plist%bII,plist%owner,ierr) 

    enddo    



    nullify(plist)


 !   call fput_mllist(par%mllistset,mllist,status)

    if (status > 2) then
      ierr = 3
    endif

 

  end subroutine format_mllist
   

!--------------------------------------------------------------------------------

  subroutine init_mllist1d(mllist)

  !------------------------------------------------------------
  ! Initialize mllist 1d table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:)     ::  mllist  

    mllist%id_src      = 0
    mllist%box_id_src  = 0
    mllist%id_inst     = 0
    mllist%id_band     = 0
    mllist%id_cluster  = 0
    mllist%scts        = 0.0
    mllist%scts_err    = 0.0
    mllist%x_ima       = 0.0
    mllist%x_ima_err   = 0.0
    mllist%y_ima       = 0.0
    mllist%y_ima_err   = 0.0
    mllist%ext         = 0.0
    mllist%ext_err     = 0.0
    mllist%like        = 0.0
    mllist%ext_like    = 0.0
    mllist%bg_map      = 0.0
    mllist%exp_map     = 0.0
    mllist%flux        = 0.0
    mllist%flux_err    = 0.0
    mllist%rate        = 0.0
    mllist%rate_err    = 0.0
    mllist%ra          = 0.0
    mllist%dec         = 0.0
    mllist%radec_err   = 0.0
    mllist%lII         = 0.0
    mllist%bII         = 0.0
    mllist%rawx        = 0.0
    mllist%rawy        = 0.0
    mllist%offax       = 0.0
    mllist%ccdnr       = 0.0
    mllist%hr1         = 0.0
    mllist%hr1_err     = 0.0
    mllist%hr2         = 0.0
    mllist%hr2_err     = 0.0
    mllist%hr3         = 0.0
    mllist%hr3_err     = 0.0
!    mllist%hr4         = 0.0
!    mllist%hr4_err     = 0.0
    mllist%cutrad      = 0.0
    mllist%maskfrac    = 0.0
    mllist%eef         = 0.0
    mllist%vignetting  = 0.0
    mllist%ontime      = 0.0
    mllist%dist_nn     = 0.0
!    mllist%flag        = ''
    mllist%ndof        = 0
    mllist%cluflag     = .false.
    mllist%scutrad     = 0.0
    mllist%ecutrad     = 0.0
    ! AGE HAD ADDED APETOOL PARAMS
    mllist%ape_cts  = 0
    mllist%ape_bkg  = 0.0
    mllist%ape_exp  = 0.0   
    mllist%ape_eef  = 0.0
    mllist%ape_radius  = 0.0       
    mllist%ape_pois  = 0.0

  end subroutine init_mllist1d

!--------------------------------------------------------------------------------
  subroutine init_mllist2d(mllist)

  !------------------------------------------------------------
  ! Initialize mllist 2d table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:,:)     ::  mllist  

    mllist%id_src      = 0
    mllist%box_id_src  = 0
    mllist%id_inst     = 0
    mllist%id_band     = 0
    mllist%id_cluster  = 0
    mllist%scts        = 0.0
    mllist%scts_err    = 0.0
    mllist%x_ima       = 0.0
    mllist%x_ima_err   = 0.0
    mllist%y_ima       = 0.0
    mllist%y_ima_err   = 0.0
    mllist%ext         = 0.0
    mllist%ext_err     = 0.0
    mllist%like        = 0.0
    mllist%ext_like    = 0.0
    mllist%bg_map      = 0.0
    mllist%exp_map     = 0.0
    mllist%flux        = 0.0
    mllist%flux_err    = 0.0
    mllist%rate        = 0.0
    mllist%rate_err    = 0.0
    mllist%ra          = 0.0
    mllist%dec         = 0.0
    mllist%radec_err   = 0.0
    mllist%lII         = 0.0
    mllist%bII         = 0.0
    mllist%rawx        = 0.0
    mllist%rawy        = 0.0
    mllist%offax       = 0.0
    mllist%ccdnr       = 0.0
    mllist%hr1         = 0.0
    mllist%hr1_err     = 0.0
    mllist%hr2         = 0.0
    mllist%hr2_err     = 0.0
    mllist%hr3         = 0.0
    mllist%hr3_err     = 0.0
!    mllist%hr4         = 0.0
!    mllist%hr4_err     = 0.0
    mllist%cutrad      = 0.0
    mllist%maskfrac    = 0.0
    mllist%eef         = 0.0
    mllist%vignetting  = 0.0
    mllist%ontime      = 0.0
    mllist%dist_nn     = 0.0
!    mllist%flag        = ''
    mllist%ndof        = 0
    mllist%cluflag     = .false.
    mllist%scutrad     = 0.0
    mllist%ecutrad     = 0.0
    ! AGE HAD ADDED APETOOL PARAMS
    mllist%ape_cts     = 0
    mllist%ape_bkg     = 0.0
    mllist%ape_exp     = 0.0   
    mllist%ape_eef     = 0.0
    mllist%ape_radius  = 0.0       
    mllist%ape_pois    = 0.0

  end subroutine init_mllist2d

!--------------------------------------------------------------------------------

  subroutine init_mllist3d(mllist)

  !------------------------------------------------------------
  ! Initialize mllist 3d table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:,:,:)     ::  mllist  

    mllist%id_src      = 0
    mllist%box_id_src  = 0
    mllist%id_inst     = 0
    mllist%id_band     = 0
    mllist%id_cluster  = 0
    mllist%scts        = 0.0
    mllist%scts_err    = 0.0
    mllist%x_ima       = 0.0
    mllist%x_ima_err   = 0.0
    mllist%y_ima       = 0.0
    mllist%y_ima_err   = 0.0
    mllist%ext         = 0.0
    mllist%ext_err     = 0.0
    mllist%like        = 0.0
    mllist%ext_like    = 0.0
    mllist%bg_map      = 0.0
    mllist%exp_map     = 0.0
    mllist%flux        = 0.0
    mllist%flux_err    = 0.0
    mllist%rate        = 0.0
    mllist%rate_err    = 0.0
    mllist%ra          = 0.0
    mllist%dec         = 0.0
    mllist%radec_err   = 0.0
    mllist%lII         = 0.0
    mllist%bII         = 0.0
    mllist%rawx        = 0.0
    mllist%rawy        = 0.0
    mllist%offax       = 0.0
    mllist%ccdnr       = 0.0
    mllist%hr1         = 0.0
    mllist%hr1_err     = 0.0
    mllist%hr2         = 0.0
    mllist%hr2_err     = 0.0
    mllist%hr3         = 0.0
    mllist%hr3_err     = 0.0
!    mllist%hr4         = 0.0
!    mllist%hr4_err     = 0.0
    mllist%cutrad      = 0.0
    mllist%maskfrac    = 0.0
    mllist%eef         = 0.0
    mllist%vignetting  = 0.0
    mllist%ontime      = 0.0
    mllist%dist_nn     = 0.0
!    mllist%flag        = ''
    mllist%ndof        = 0
    mllist%cluflag     = .false.
    mllist%scutrad     = 0.0
    mllist%ecutrad     = 0.0
    ! AGE HAD ADDED APETOOL PARAMS
    mllist%ape_cts    = 0
    mllist%ape_bkg    = 0.0
    mllist%ape_exp    = 0.0   
    mllist%ape_eef    = 0.0
    mllist%ape_radius = 0.0       
    mllist%ape_pois   = 0.0
  end subroutine init_mllist3d

!--------------------------------------------------------------------------------




  subroutine init_boxlist1d(boxlist)

  !------------------------------------------------------------
  ! Initialize boxlist table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:)     ::  boxlist
    
    boxlist%id_src = 0     
    boxlist%id_inst = 0        
    boxlist%id_band = 0       
    boxlist%scts = 0.0         
    boxlist%scts_err = 0.0       
    boxlist%box_cts = 0.0        
    boxlist%x_ima  = 0.0         
    boxlist%x_ima_err = 0.0      
    boxlist%y_ima  = 0.0         
    boxlist%y_ima_err = 0.0      
    boxlist%like  = 0.0          
    boxlist%bg_map  = 0.0        
    boxlist%bg_raw  = 0.0        
    boxlist%exp_map = 0.0        
    boxlist%flux  = 0.0          
    boxlist%flux_err = 0.0       
    boxlist%rate  = 0.0          
    boxlist%rate_err = 0.0       
    boxlist%ra  = 0.0            
    boxlist%dec = 0.0            
    boxlist%radec_err = 0.0
    boxlist%lII = 0.0            
    boxlist%bII = 0.0            
    boxlist%hr1 =      -999.9         
    boxlist%hr1_err =  -999.9       
    boxlist%hr2  =     -999.9       
    boxlist%hr2_err =  -999.9     
    boxlist%hr3  =     -999.9        
    boxlist%hr3_err  =  -999.9    
    boxlist%vignetting = 0.0     
    boxlist%box_size = 0.0       
    boxlist%eef  = 0.0           
    boxlist%dist_nn  = 0.0       
    boxlist%maskzero = 0 
    boxlist%nzero = 0   
    boxlist%ndof = 0        


  end subroutine init_boxlist1d
   
!--------------------------------------------------------------------------------   
 

  subroutine init_boxlist2d(boxlist)

  !------------------------------------------------------------
  ! Initialize boxlist table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:,:)     ::  boxlist
    
    boxlist%id_src = 0     
    boxlist%id_inst = 0        
    boxlist%id_band = 0       
    boxlist%scts = 0.0         
    boxlist%scts_err = 0.0       
    boxlist%box_cts = 0.0        
    boxlist%x_ima  = 0.0         
    boxlist%x_ima_err = 0.0      
    boxlist%y_ima  = 0.0         
    boxlist%y_ima_err = 0.0      
    boxlist%like  = 0.0          
    boxlist%bg_map  = 0.0        
    boxlist%bg_raw  = 0.0        
    boxlist%exp_map = 0.0        
    boxlist%flux  = 0.0          
    boxlist%flux_err = 0.0       
    boxlist%rate  = 0.0          
    boxlist%rate_err = 0.0       
    boxlist%ra  = 0.0            
    boxlist%dec = 0.0            
    boxlist%radec_err = 0.0      
    boxlist%lII = 0.0            
    boxlist%bII = 0.0            
    boxlist%hr1 =       -999.9           
    boxlist%hr1_err  =  -999.9       
    boxlist%hr2  =      -999.9        
    boxlist%hr2_err  =  -999.9       
    boxlist%hr3  =      -999.9       
    boxlist%hr3_err  = 0.0       
    boxlist%vignetting = 0.0     
    boxlist%box_size = 0.0       
    boxlist%eef  = 0.0           
    boxlist%dist_nn  = 0.0       
    boxlist%maskzero = 0   
    boxlist%nzero = 0   
    boxlist%ndof = 0        


  end subroutine init_boxlist2d
   
!--------------------------------------------------------------------------------   
 
 subroutine init_boxlist3d(boxlist)

  !------------------------------------------------------------
  ! Initialize boxlist table elements with zero
  !------------------------------------------------------------   
  
    type(srcrow_type),dimension(:,:,:)     ::  boxlist
    
    boxlist%id_src = 0     
    boxlist%id_inst = 0        
    boxlist%id_band = 0       
    boxlist%scts = 0.0         
    boxlist%scts_err = 0.0       
    boxlist%box_cts = 0.0        
    boxlist%x_ima  = 0.0         
    boxlist%x_ima_err = 0.0      
    boxlist%y_ima  = 0.0         
    boxlist%y_ima_err = 0.0      
    boxlist%like  = 0.0          
    boxlist%bg_map  = 0.0        
    boxlist%bg_raw  = 0.0        
    boxlist%exp_map = 0.0        
    boxlist%flux  = 0.0          
    boxlist%flux_err = 0.0       
    boxlist%rate  = 0.0          
    boxlist%rate_err = 0.0       
    boxlist%ra  = 0.0            
    boxlist%dec = 0.0            
    boxlist%radec_err = 0.0      
    boxlist%lII = 0.0            
    boxlist%bII = 0.0            
    boxlist%hr1 = 0.0            
    boxlist%hr1_err  = -999.9      
    boxlist%hr2  =    -999.9        
    boxlist%hr2_err  =  -999.9       
    boxlist%hr3  =    -999.        
    boxlist%hr3_err  =  -999.      
    boxlist%vignetting = 0.0     
    boxlist%box_size = 0.0       
    boxlist%eef  = 0.0           
    boxlist%dist_nn  = 0.0       
    boxlist%maskzero = 0   
    boxlist%nzero = 0   
    boxlist%ndof = 0        


  end subroutine init_boxlist3d
   


  subroutine addval( value1, value2, add_val, ierr)

    
    real(kind=real32), intent(in) :: value1, value2
    real(kind=real32), intent(out) :: add_val
    integer(kind=int32), intent(out) :: ierr
    character(len=*),parameter       :: rtname = 'ADDVAL'

    ierr=0

    if(.not. value1 >= 0 ) then
       add_val = value2
    else
       if(.not. value2 >= 0 ) then
          add_val = value1
       else
          add_val = value1 + value2 
       end if
    end if

  end subroutine addval

!--------------------------------------------------------------------------------


  subroutine addwitherr( value1, error1, value2, error2, add_val, add_err, ierr)

  !--------------------------------------------------------------------- 
  ! Add 2 values  and calculate
  ! statistical error.
  ! Use only input values and errors > 0.0    
  !---------------------------------------------------------------------   
 
    real(kind=real32), intent(in) :: value1, value2, error1,  error2
    real(kind=real32), intent(out) :: add_val, add_err
    integer(kind=int32), intent(out) :: ierr
    character(len=*),parameter       :: rtname = 'ADDWITHERR'

    ierr=0

    if(.not. value1 > 0 ) then
       add_val = value2
       add_err = error2
    else
       if(.not. value2 > 0 ) then
          add_val = value1
          add_err = error1
       else
          add_val = value1 + value2
          add_err = 0.0
          if (error1 > 0 .and. error2 > 0) then
            add_err = (error1**2 + error2**2)**0.5
          else if (error1 > 0) then
             add_err = error1
          else if (error2 > 0) then
             add_err = error2
          endif   
       end if
    end if

  end subroutine addwitherr

!--------------------------------------------------------------------------------

  subroutine average( value1, value2, av_val, ierr)


  !--------------------------------------------------------------------- 
  ! Calculate simple average of value1, value2 .
  ! Use only input values  > 0.0.  
  !---------------------------------------------------------------------   

    real(kind=real32), intent(in)   :: value1, value2
    real(kind=real32), intent(out)   :: av_val
    integer(kind=int32), intent(out) :: ierr
    real(kind=real32)                :: newval
    character(len=*),parameter       :: rtname = 'AVERAGE'

    ierr=0

    if(.not. value1 > 0 ) then
       newval = value2
    else
      if(.not. value2 > 0 ) then
        newval = value1
      else
        newval = (value1 + value2)/2.0
      endif
    endif
    av_val = newval

  end subroutine average

!--------------------------------------------------------------------------------



  subroutine w_average( value1, error1, value2, error2, av_val, av_err, ierr)

  !---------------------------------------------------------------------    
  ! Calculate weighted mean with eroor from  value1, value2 using
  ! 1/error1^2, 1/error2^2 as weights. Input/output values real32.
  !---------------------------------------------------------------------    



    real(kind=real32), intent(in)    :: value1, error1, value2, error2
    real(kind=real32), intent(out)   :: av_val, av_err
    integer(kind=int32), intent(out) :: ierr
    real(kind=real32)                :: w1, w2, totw, newval, newerr ! weights
    character(len=*),parameter       :: rtname = 'W_AVERAGE'


    ierr=0

    if(.not. error1 > 0.0 ) then
       av_val = value2
       av_err = error2
    else
       if(.not. error2 > 0.0 ) then
          av_val = value1
          av_err = error1
       else
          ! error1 /= 0 && error2 /= 0: weighted average
          w1 =  1.0/error1**2
          w2 =  1.0/error2**2
          totw = w1 + w2
          newval = (w1 * value1 + w2 * value2)/totw
          newerr = 1.0/sqrt( totw )
          av_val = newval
          av_err = newerr
       end if
    end if

  end subroutine w_average

!--------------------------------------------------------------------------------

  subroutine dw_average( value1, error1, value2, error2, av_val, av_err, ierr)

  !---------------------------------------------------------------------    
  ! Calculate weighted mean with eroor from  value1, value2 using
  ! 1/error1^2, 1/error2^2 as weights.Input/output values real64.
  !---------------------------------------------------------------------    

    real(kind=real64), intent(in)    :: value1, error1, value2, error2
    real(kind=real64), intent(out)   :: av_val, av_err
    integer(kind=int32), intent(out) :: ierr
    real(kind=real64)                :: w1, w2, totw, newval, newerr ! weights
    character(len=*),parameter       :: rtname = 'DW_AVERAGE'

    ierr=0

    if(.not. error1 > 0.0 ) then
       av_val = value2
       av_err = error2
    else
       if(.not. error2 > 0.0 ) then
          av_val = value1
          av_err = error1
       else
          ! error1 /= 0 && error2 /= 0: weighted average
          w1 =  1.0/error1**2
          w2 =  1.0/error2**2
          totw = w1 + w2
          newval = (w1 * value1 + w2 * value2)/totw
          newerr = 1.0/sqrt( totw )
          av_val = newval
          av_err = newerr
       end if
    end if

  end subroutine dw_average

!--------------------------------------------------------------------------------

  subroutine hrcalc( rate1, error1, rate2, error2, hr, hr_err,ierr)

  !---------------------------------------------------------------------    
  ! Calculate hardness ratio with errors between  
  ! rate1 (soft) and rate2 (hard).
  ! 
  !---------------------------------------------------------------------    

    real(kind=real32), intent(in)    :: rate1, error1, rate2, error2
    real(kind=real32), intent(out)   :: hr, hr_err
    integer(kind=int32), intent(out) :: ierr
    character(len=*),parameter       :: rtname = 'HRCALC'

    ierr=0 

      if (rate1 <= 0.0 .and.  rate2 <= 0.0) then
        hr=-999.9
        ierr=1
      else
        hr=(max(0.,rate2)-max(0.,rate1))/ &
                 (max(0.,rate1)+max(0.,rate2))
        if (error1 <= 0.0 .and.  error2 <= 0.0) then
          hr_err=-999.9
          ierr=1
        else
          hr_err=2.*sqrt( (rate1*error2)**2 + (rate2*error1)**2) /  &
                (max(0.,rate1)+max(0.,rate2))**2
        endif
      endif

  end subroutine hrcalc

!-------------------------------------------------------------------------------


  real(kind=real32) function gammq(a,x)


  !--------------------------------------------------------------------
  !  returns the negative logarithm of the 
  !  complemente of the incomplete gamma function p(a,x) 
  !
  !  author : s.doebereiner date: 11-sep-1996
  !  update : G. Lamer      date: 02-05-2011 
  !                         adapt to eROSITA SASS 
  !
  !  returns the incomplete gamma function p(a,x) 
  !  in the maximum-likelihood representation -ln(p) 
  !  a and x must be >0 !
  !  modified routines gcf and gser from press et al., numerical recipes, 
  !   6.2, p.162
  !  a                   r4  i   argument
  !  x                   r4  i   argument 
  !  gammq               r4  o   -ln(q(a,x))
  !
  !  gammln              sr      returns ln of the gamma function  
  !  wrerr               sr      write error message
  !--------------------------------------------------------------------

      real(kind=real32)          ::  a,x
      real(kind=real32),parameter ::  eps=3.E-7
      real(kind=real32)          ::  gln,gold
      real(kind=real32)          ::  a0,a1,b0,b1,fac,an,ana,anf,g
      real(kind=real32)          ::  ap,sum,del,gamser,gammcf
      integer(kind=int32)        ::  n,ierr     
      integer(kind=int32)        ::  itmax=1000
      character(len=*),parameter ::  rtname='GAMMQ'


      ierr = 0
      gammq = 0.
    
      x = max(x,eps) 


      if (x.lt.a+1) then
        gln=gammln(a)
        ap=a
        sum=1./a
        del=sum
        do n=1,itmax
          ap=ap+1.
          del=del*x/ap
          sum=sum+del
          if (abs(del).lt.abs(sum)*eps) then 
            exit
          endif
        end do
        if (n >= itmax) then 
          ierr = 1 
          call wrchatty(rtname,'Warning: a too large, itmax too small ')
!          write(*,*) '1: gammq a,x= ',a,x
        endif

        gamser = sum * exp( -x + a*log(x) - gln)
        gammq = -alog(1.0-gamser)
        
      else

        gln=gammln(a)
        gold=0.
        a0=1.
        a1=x
        b0=0.
        b1=1.
        fac=1.

        do n=1,itmax
          an=float(n)
          ana=an-a
          a0=(a1+a0*ana)*fac
          b0=(b1+b0*ana)*fac
          anf=an*fac
          a1=x*a0+anf*a1
          b1=x*b0+anf*b1
          if (a1.ne.0) then
            fac=1./a1
            g=b1*fac
            if (abs((g-gold)/g).lt.eps) then 
              exit
            endif
            gold=g
          endif
        enddo
        if (n >= itmax) then 
          ierr = 1 
          call wrchatty(rtname,'Warning: a too large, itmax too small ')
!          write(*,*) '2: gammq a,x= ',a,x
        endif

        gammq = x - a*alog(x) + gln - alog(g)

      endif 

      end function gammq

!-------------------------------------------------------------------------------

  real(kind=real32) function gammp (a,x)

  !------------------------------------------------------------------
  !  returns the negative logarithm of the 
  !  incomplete gamma function p(a,x) 
  !  author : s.doebereiner date: 11-sep-1996
  !  update : G. Lamer      date: 02-05-2011 
  !                         adapt to eROSITA SASS 
  !  returns the incomplete gamma function p(a,x) 
  !  in the maximum-likelihood representation -ln(p) 
  !  a and x must be >0 !
  !  modified routines gcf and gser from press et al., numerical recipes, 
  !   6.2, p.162
  !
  !  a                   r4  i   argument
  !  x                   r4  i   argument 
  !  gammp               r4    o -ln(p(a,x))
  !
  !  gammln              sr      returns ln of the gamma function  
  !  wrerr               sr      write error message
  !------------------------------------------------------------------


      real(kind=real32)          ::    a,x
      real(kind=real32)          ::    eps=3.E-7
      real(kind=real32)          ::    gln,gold
      real(kind=real32)          ::    a0,a1,b0,b1,fac,an,ana,anf,g
      real(kind=real32)          ::    ap,sum,del,gamser,gammcf
      integer(kind=int32)        ::    itmax=1000
      integer(kind=int32)        ::    n,ierr
      character(len=*),parameter ::    rtname='GAMMP'

      ierr = 0
      gammp = 0.

      if (x.lt.a+1) then
        gln=gammln(a)
        ap=a
        sum=1./a
        del=sum
        do n=1,itmax
          ap=ap+1.
          del=del*x/ap
          sum=sum+del
          if (abs(del).lt.abs(sum)*eps) then
            exit
          endif
        enddo 
        if (n >= itmax) then
          ierr = 1 
          write(*,*) 'gammp',a,x,itmax,sum
          call wrchatty(rtname,'Warning: a too large, itmax too small ')
        endif

        gammp = -alog(sum) + x - a*alog(x) + gln      
      else
        gln=gammln(a)
        gold=0.
        a0=1.
        a1=x
        b0=0.
        b1=1.
        fac=1.

        do n=1,itmax
          an=float(n)
          ana=an-a
          a0=(a1+a0*ana)*fac
          b0=(b1+b0*ana)*fac
          anf=an*fac
          a1=x*a0+anf*a1
          b1=x*b0+anf*b1
          if (a1.ne.0) then
            fac=1./a1
            g=b1*fac
            if (abs((g-gold)/g).lt.eps) then
              exit
            endif
            gold=g
          endif
        enddo
        if (n >= itmax) then 
          ierr = 1 
          write(*,*) 'gammp',a,x,itmax
          call wrchatty(rtname,'Warning: a too large, itmax too small ')
        endif

        gammcf = exp(-x + a*alog(x) - gln) * g
        gammp = -alog(1.-gammcf)

      endif 


      end function gammp

!-------------------------------------------------------------------------------


  real(kind=real32) function gammln (xx)

  !----------------------------------------------------------------
  !  returns natural logarithm of the gamma function for xx>=1
  !  author : s.doebereiner date: 11-sep-1996
  !  update : G. Lamer      date: 02-05-2011 
  !                         adapt to eROSITA SASS 
  !  returns natural logarithm of the gamma function for xx>=1
  !  routine gammln from press et al., numerical recipes, 6.1, p.157
  !
  !  xx                  r4  i   argument
  !  gammln              r4    o ln(gamma(xx))
  !
  !  wrerr               sr      write error message
  !----------------------------------------------------------------


      real(kind=real32)               ::    xx 
      real(kind=real64),dimension(6)  :: cof =        &
                                       (/76.18009173D0,-86.50532033D0,24.01409822D0,&
                                       -1.231739516D0,.120858003d-2,-.536382D-5 /)
      real(kind=real64)               :: stp=2.50662827465D0
      real(kind=real64)               :: x,tmp,ser
      integer                         :: j,ierr
      character(len=*),parameter      ::  rtname='GAMMLN'


      ierr = 0
      gammln = 0.

      if (xx.le.0.0) then
        ierr = 1
        call wrerr(rtname,ierr,'argument <=0 not allowed ') 
      endif

      if (ierr.eq.0) then 

        ! if xx<1 then use reflection formula (see press et al., eq. 6.1.4)     
        if (xx.lt.1) then
          x = xx
        else
          x = xx - 1.0D0
        endif       

        tmp=x+5.5D0
        tmp=(x+0.5D0)*log(tmp)-tmp
        ser=1.0D0
        do  j=1,6
          x=x+1.0D0
          ser=ser+cof(j)/x
        enddo

      gammln=tmp+log(stp*ser)
      if (xx.lt.1) gammln=gammln-log(xx)

      endif

      end function gammln


!-------------------------------------------------------------------------------

  integer function ixb(ix,im,ncom)

  !---------------------------------------------------------------------
  ! Subtract offset ixb0(im) from pixel coordinate ix
  ! IM: number of detection image 
  ! NCOM : image rebinning factor
  ! Use global variables:  XIMBUF and IXB0, COMPRESS_FLAG
  ! COMPRESS_FLAG : = true for memory saving image reading (par%compress_flag)
  !---------------------------------------------------------------------

  integer(kind=int32) :: ix, im
  
  if (compress_flag) then
    if (ncom == 1) then
      ixb = max(min(ix-ixb0(im),ximbuf),1)
    else
      ixb = max(min(ix-ixb0(im)/ncom,ximbuf),1)
    endif
  else 
    ixb=ix
  endif
  

  end function ixb

!-------------------------------------------------------------------------------

  integer function iyb(iy,im,ncom)

  !---------------------------------------------------------------------
  ! Subtract offset iyb0(im) from pixel coordinate iy
  ! NCOM : image rebinning factor
  ! FLAG : = true for memory saving image reading (par%compress_flag)
  ! Use global variables:  XIMBUF and IXB0
  ! 
  !---------------------------------------------------------------------

  integer(kind=int32) :: iy, im

  if (compress_flag) then  
    if (ncom == 1) then 
      iyb = max(min(iy-iyb0(im),yimbuf),1)
    else
      iyb = max(min(iy-iyb0(im)/ncom,yimbuf),1)
    endif
  else  
    iyb=iy
  endif

  end function iyb


!--------------------------------------------------------------------------------

  integer function ixm(ix,im)

  !---------------------------------------------------------------------
  !  Subtract offset ixb0(im) from pixel coordinate ix.
  !  Similar to function ixb, but simplified for use in erml.
  !---------------------------------------------------------------------

  integer :: ix, im

  ixm = max(min(ix-ixb0(im),ximbuf),1)

  end function ixm

!--------------------------------------------------------------------------------

  integer function iym(iy,im)

  !---------------------------------------------------------------------
  !  Subtract offset iyb0(im) from pixel coordinate iy.
  !  Similar to function iyb, but simplified for use in erml.
  !---------------------------------------------------------------------

  integer :: iy, im

  iym = max(min(iy-iyb0(im),yimbuf),1)

  end function iym

!--------------------------------------------------------------------------------


  character(len=clong) function lowcase(word)

  !---------------------------------------------------
  !convert a word to lower case
  !---------------------------------------------------

  character(len=*)      :: word
!  character(len=clong)  :: lowcase
  integer               :: i,ic,nlen

  nlen = len(word)
  lowcase = word
  do i=1,nlen
    ic = ichar(word(i:i))
    if (ic >= 65 .and. ic < 90)  lowcase(i:i) = char(ic+32)
  end do 

  end function lowcase

!---------------------------------------------------------------------------------------


  subroutine calc_sourceowner(ra,dec,srcrad,glon,glat,flag,ierr) 

 
    !--------------------------------------------------------------------
    ! Determines proprietary sky area (IKI, MPE or MPE_IKI from 
    ! RA, DEC [degrees], and sets  bits 0 and 1 of FLAG accordingly    
    ! NOTE: owner=MPE=> bit0=0, owner=IKI=> bit1=0, MPE+IKI: both bits =0. 
    ! SRCRAD (in deg) specifies a cone radius around (ra,dec). If the cone
    ! overlaps with the exclusive area of the respective other owner, the owner 
    ! flag is not set.
    ! SRCRAD must be smaller than the half width of the dual ownership stripe
    ! (currently 19 arcmin), otherwise  an error  message and flag=3 (no owner)
    ! will result.   
    !--------------------------------------------------------------------


    implicit none
 
    real(kind=real64)                  :: ra    
    real(kind=real64)                  :: dec
    real(kind=real64)                  :: srcrad
    real(kind=real64)                  :: glon
    real(kind=real64)                  :: glat
    integer(kind=int8)                 :: flag
    integer(kind=int32)                :: ierr

    real(kind=real64)                  :: ra_rad, dec_rad    
    integer(kind=int32)                :: status
    real(kind=real64)                  :: dlon_dual,sin_hw_dual,hw_dual_eff
    logical(kind=log32)                :: ikiflag,mpeflag
    real(kind=real64),parameter        :: hw_dual_deg=0.316666666667 ! half width of dual ownership stripe = 19 arcmin
    real(kind=real64),parameter        :: dpi=3.14159265358979312,d2pi=6.28318530717958623
    
    character(len=*),parameter         :: rtname='CALC_OWNERFLAG'  


    ierr=0
    status=0
    ikiflag=.false.
    mpeflag=.false.
    flag=3


    hw_dual_eff=hw_dual_deg-srcrad

    if  (hw_dual_eff < 0.0) then
      ierr=3
      call WRERR(rtname,ierr,'srcrad value too large')
    else

      sin_hw_dual=sin(hw_dual_eff*deg2rad)

      ! catch ra outside [0,360]:
      ra_rad=dradrs(ra*deg2rad)
      dec_rad=dec*deg2rad

      call equgal(2451544.5d0,.true.,ra_rad, dec_rad, glon, glat, status)

      if (status == 0) then

        if (sin_hw_dual/cos(glat) < 1.D0) then 
          if (glon < 0.D0) glon = glon+d2pi
          dlon_dual =  asin(sin_hw_dual/cos(glat))    
          if (glon  >= d2pi-dlon_dual .or.  (glon >= 0d0    .and. glon <= dpi+dlon_dual)) ikiflag=.true.
          if ((glon  >= dpi -dlon_dual .and. glon <= d2pi)  .or.  glon <= dlon_dual)      mpeflag=.true.
        else
          ! source is close to galactic poles and within dual OS stripe:
          ikiflag=.true.
          mpeflag=.true.
        endif

        flag = ior(flag,int(3,kind=1))
        if (mpeflag) flag = ibclr(flag,0)
        if (ikiflag) flag = ibclr(flag,1)

        glon=glon/deg2rad
        glat=glat/deg2rad

      else
        ierr=3
        call wrerr(rtname,ierr,'Error getting galactic coordinates')
      endif
    endif
  

  end subroutine calc_sourceowner




end module detect_io


    

     

    


    
      
  
