! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module quaternions

  !----------------------------------------------------------------
  ! eROSITA SASS quaternion library
  ! Author: GLamer, AIP
  !
  ! All routines assume the following notation:
  ! q%arr(1)=qx  
  ! q%arr(2)=qy  
  ! q%arr(3)=qz   
  ! q%arr(4)=qw 
  ! 
  !----------------------------------------------------------------

  use iso_fortran_env
  use param

  implicit none


  type quaternion_type
  !--------------------------------------------------------------------
  ! Changed matrix and att from 'pointer' to fixed size arrays,
  ! since Fortran standard does not allow arrays of pointers 
  !--------------------------------------------------------------------     

    real(kind=real64),dimension(4)              ::   arr
!    real(kind=real64),pointer,dimension(:)     ::   v
!    real(kind=real64),pointer                  ::   w 
!    real(kind=real64),pointer,dimension(:,:)   ::   matrix => NULL()
!    real(kind=real64),pointer,dimension(:)     ::   att => NULL()
    real(kind=real64),dimension(3,3)   ::   matrix
    real(kind=real64),dimension(3)     ::   att
    contains  

    procedure :: QcalcMatrix
    procedure :: QfromMatrix
    procedure :: QfromAtt
    procedure :: QcalcAtt
    procedure :: QcalcQuat
    procedure :: QcalcAtt1
    procedure :: QcalcAtt2
    procedure :: QrotVector
    procedure :: Qnorm
    procedure :: Qpositive   
 

    procedure :: calcMatrix => QcalcMatrix
    procedure :: fromMatrix => QfromMatrix
    procedure :: fromAtt    => QfromAtt

!    procedure :: calcAtt    => QcalcAtt2
    procedure :: calcAtt    => QcalcAtt1
!    procedure :: calcAtt    => QcalcAtt

    procedure :: calcQuat   => QcalcQuat

    procedure :: rotVector  => QrotVector
    procedure :: normalize  => Qnorm
    procedure :: positive  =>  Qpositive


  end type quaternion_type




!  interface operator(*)
!    module procedure multiplyQuaternions
!  end interface

  
!  interface operator(=)
!    module procedure copyQuaternions
!  end interface


contains


!---------------------------------------------------------

  function dot4d(q1,q2)


    real(kind=real64),dimension(4)          :: q1,q2
    real(kind=real64)                       :: dot4d

    dot4d = q1(1)*q2(1) +  q1(2)*q2(2) +   q1(3)*q2(3) +  q1(4)*q2(4)

  end function dot4d 

!---------------------------------------------------------



!---------------------------------------------------------

  function dot3d(u,v)

    implicit none
    real(kind=real64),dimension(3) ::   u,v
    real(kind=real64)              ::   dot3d

    dot3d = u(1)*u(1) +  u(2)*u(2) +  u(3)*u(3)

  end function dot3d   

!---------------------------------------------------------

  function cross3d(u,v)

    implicit none
    real(kind=real64),dimension(3) ::   u,v
    real(kind=real64),dimension(3) ::   cross3d

    cross3d(1) =  u(2)*v(3)  - u(3)*v(2) 
    cross3d(2) =  u(3)*v(1)  - u(1)*v(1)  
    cross3d(3) =  u(1)*v(2)  - u(2)*v(1)  

  end function cross3d  


!---------------------------------------------------------

  function Qinv(quat)

    type (quaternion_type)               :: quat
    type (quaternion_type)               :: Qinv  

    
    Qinv%arr(1:3) =    quat%arr(1:3)
    Qinv%arr(4)   =  - quat%arr(4) 
    if (Qinv%arr(4) < 0.D0) Qinv%arr =  -Qinv%arr

  end function Qinv


!---------------------------------------------------------
 

  subroutine Qpositive(quat)
   
  !--------------------------------------------------------
  ! If real value of quaternion is negative, change sign
  ! of all elements.
  !--------------------------------------------------------

    class (quaternion_type)              :: quat


    if (quat%arr(4) < 0.D0) quat%arr =  -quat%arr


  end subroutine Qpositive

!---------------------------------------------------------
 


 
  function Qdiv(quat1,quat2)
  
  !---------------------------------
  ! Division of quat1 by quat2
  ! Qdiv =  quat1 * quat2^-1
  !---------------------------------

 
    type (quaternion_type)               :: quat1,quat2
    type (quaternion_type)               :: Qdiv  
 

    Qdiv = Qmult(quat1, Qinv(quat2))
     

  end function Qdiv

!---------------------------------------------------------

  function Qmult(quat1,quat2)
 
    type (quaternion_type)               :: quat1,quat2
    type (quaternion_type)               :: Qmult
    real(kind=real64), dimension(4)      :: q1,q2   
    

      q1 = quat1%arr
      q2 = quat2%arr   


      Qmult%arr(1) =  q1(4)*q2(1) - q1(3)*q2(2) + q1(2)*q2(3) + q1(1)*q2(4)
      Qmult%arr(2) =  q1(3)*q2(1) + q1(4)*q2(2) - q1(1)*q2(3) + q1(2)*q2(4)
      Qmult%arr(3) = -q1(2)*q2(1) + q1(1)*q2(2) + q1(4)*q2(3) + q1(3)*q2(4)
      Qmult%arr(4) = -q1(1)* q2(1) - q1(2)*q2(2) - q1(3)*q2(3) + q1(4)*q2(4)


  end function Qmult  

!---------------------------------------------------------


  subroutine  QrotVector(quat,v,vrot,ierr)


    
    class (quaternion_type)               :: quat
    real(kind=real64),dimension(3)        :: v
    real(kind=real64),dimension(3)        :: vrot
    integer(kind=int32)                   :: ierr
    
    ierr=0

!    if (associated(quat%matrix)) then
      vrot = matmul(quat%matrix,v)
!    else
!      ierr=3
!    endif
    

  end subroutine QrotVector

!------------------------------------------------------------------------------------------------------------------  

   subroutine  Qnorm(quat,ierr)
     

     class (quaternion_type)               :: quat 
     integer(kind=int32)                   :: ierr
     real(kind=real64)                     :: length2
     character(len=*),parameter            :: rtname='Qnorm'
 
     ierr=0

     length2 = quat%arr(1)*quat%arr(1)+quat%arr(2)*quat%arr(2)+ quat%arr(3)*quat%arr(3)+quat%arr(4)*quat%arr(4)
 
     if (length2 == 0.D0) then
       ierr=3
       call wrerr(rtname,ierr,'Error: zero length quaternion')
     else if (length2 .ne. 1.D0) then
       quat%arr = quat%arr / sqrt(length2)  
     endif
     if  (quat%arr(4) < 0.D0)  quat%arr = -quat%arr

   end subroutine Qnorm



!------------------------------------------------------------------------------------------------------------------  


  subroutine  Qslerp(q1,q2,t1,t2,tint,qint,ierr)

    !---------------------------------------------------------
    ! Spherical linear interpolation  between q1,q2 at
    ! time stamps t1,t2 to q_int at time t_int. 
    !--------------------------------------------------------- 


    type(quaternion_type)            :: q1
    type(quaternion_type)            :: q2
    real(kind=real64)                :: t1
    real(kind=real64)                :: t2
    real(kind=real64)                :: tint
    type(quaternion_type)            :: qint
    integer(kind=int32)              :: ierr
   
    type(quaternion_type)            :: qtemp 
    real(kind=real64),parameter      :: pi=3.141592653589793D0
    real(kind=real64)                :: ttemp,tn,dot12,theta
    integer(kind=int32)              :: i
    real(kind=real64)                :: f1,f2,f3
    character(len=*),parameter       :: rtname='Qslerp'
     real(kind=real64)               :: deg2rad

     deg2rad=pi/180.D0 

    ierr=0

    if (t1 == t2) then
      ierr=3
      call wrerr(rtname,ierr,'Error: t1=t2, cannot interpolate')
    else
      if (abs(tint-t1) > abs(t2-t1) .or. sign((tint-t1),(t2-t1)) < 0.D0 ) then
        ierr=2
        call wrerr(rtname,ierr,'Warning: t_int outside interval [t1,t2]')
      endif 
      if (q1%arr(4) < 0.D0) q1%arr = -q1%arr
      if (q2%arr(4) < 0.D0) q2%arr = -q2%arr

      tn=(tint-t1)/(t2-t1)

      ! Calculate half rotation angle theta: q1*q2=cos(theta)

      dot12 = dot4d(q1%arr,q2%arr)
      
      
      theta = acos(dot12)
!      write(*,*) 'theta=', dot12, theta/deg2rad
      qtemp=q2
      if (dot12 < 0.D0) then 
        q2%arr=-q2%arr
        dot12 = dot4d(q1%arr,q2%arr)
      endif  
     

      ! Calculate SLERP:
      ! q_int= (q1*sin((1-t)*theta) + q2*sin(t*theta)) / sin(theta) 

      f1=sin((1-tn)*theta)
      f2=sin(tn*theta)
      f3=sin(theta)

      !write(*,*) rtname, theta, f1,f2,f3

      do i=1,4
        qint%arr(i) = (q1%arr(i)*f1 + q2%arr(i)*f2)/f3 
      enddo

      call qint%normalize(ierr)

    endif

  end subroutine  Qslerp





!------------------------------------------------------------------------------------------------------------------  

  subroutine  QlinInt(q1,q2,t1,t2,tint,qint,ierr)

    !---------------------------------------------------------
    ! Interpolate between attitude quaternions quat1,quat2 with 
    ! time stamps time1,time2 to q_int at time t_int. 
    ! Use simple linear interpolation between quaternion elements.
    ! 
    !--------------------------------------------------------- 

    type(quaternion_type)            :: q1
    type(quaternion_type)            :: q2
    real(kind=real64)                :: t1
    real(kind=real64)                :: t2
    real(kind=real64)                :: tint
    type(quaternion_type)            :: qint
    integer(kind=int32)              :: ierr
        
    real(kind=real64)                :: ttemp,tn
    integer(kind=int32)              :: i

    character(len=*),parameter       :: rtname='QlinInt'

    ierr=0


    if (t1 == t2) then
      ierr=3
      call wrerr(rtname,ierr,'Error: t1=t2, cannot interpolate')
    else
      if (abs(tint-t1) > abs(t2-t1) .or. sign((tint-t1),(t2-t1)) < 0.D0) then
        ierr=2
        call wrerr(rtname,ierr,'Warning: t_int outside interval [t1,t2]')
      endif 
!      if (q1%arr(4) < 0.D0) q1%arr = -q1%arr
!      if (q2%arr(4) < 0.D0) q2%arr = -q2%arr
      
       tn=(tint-t1)/(t2-t1)

      do i=1,4
        qint%arr(i)=q1%arr(i)+tn*(q2%arr(i)-q1%arr(i))
      enddo   
       
      call qint%normalize(ierr)

    endif  
      
  end subroutine  QlinInt
   

    !-------------------------------------------------------------------------------------------

    subroutine  QslAve(qarr,weights,nq,qresult,ierr)


      type (quaternion_type),dimension(:)   :: qarr
      real(kind=real32), dimension(:)       :: weights
      integer(kind=int32)                   :: nq
      type (quaternion_type)                :: qresult
      integer(kind=int32)                   :: ierr

      real(kind=real64)                     :: t1,t2,tint
      integer(kind=int32)                   :: iq,status
      type (quaternion_type)                :: qsum
      real(kind=real64)                     :: wsum
      character(len=*),parameter            :: rtname='QslAve'


      ierr=0  
      status=0 
      qsum%arr= (/ 0.D0, 0.D0, 0.D0, 0.D0 /) 
      wsum=0.D0


      t1=0.D0
      t2=1.D0

      

    end subroutine QslAve




      

    !-------------------------------------------------------------------------------------------

    subroutine  Qaverage(qarr,weights,nq,qresult,ierr)
      
    !---------------------------------------------------------
    ! Average quaternions in array qarr with weights in array
    ! weights. 
    ! Only accurate if input quaternions represent positions 
    ! within close distances.
    ! 
    !--------------------------------------------------------- 


      type (quaternion_type),dimension(:)   :: qarr
      real(kind=real32), dimension(:)       :: weights
      integer(kind=int32)                   :: nq
      type (quaternion_type)                :: qresult
      integer(kind=int32)                   :: ierr

      integer(kind=int32)                   :: iq,status
      type (quaternion_type)                :: qsum
      real(kind=real64)                     :: wsum
      character(len=*),parameter            :: rtname='Qaverage'



      ierr=0  
      status=0 
      qsum%arr = (/ 0.D0, 0.D0, 0.D0, 0.D0 /) 
      wsum=0.D0

      if (size(weights) < nq .or. size(qarr) < nq) then
        ierr=3
        call wrerr(rtname,ierr,'Error: sizes of arrays qarr or weights < nq' )
      else   
        do iq=1,nq
          if (iq == 1) then
             qsum%arr=qarr(1)%arr*weights(1)
          else
            if ( dot4d(qarr(iq)%arr,qsum%arr) < 0.D0 ) qarr(iq)%arr = -qarr(iq)%arr
            qsum%arr = qsum%arr +  qarr(iq)%arr*weights(iq)
          endif 
        enddo
      endif   

      qresult=qsum
      call qresult%normalize(status)
                        

    end subroutine Qaverage


    !-------------------------------------------------------------------------------------------

    subroutine  QcalcMatrix(quat)

    !-----------------------------------------------------
    ! Calculate rotation matrix quat%matrix from 
    ! x,y,z,w array quat%arr
    !----------------------------------------------------

    implicit none
    
    class (quaternion_type)  :: quat
    integer(kind=int32)     :: ierr
    real(kind=real64)       :: xx2, yy2, zz2, xy2, wz2, zx2, wy2, yz2, wx2

    ierr=0


  !  if (.not. associated(quat%matrix))  allocate(quat%matrix(3,3))
    quat%matrix(:,:)=0.D0
   

    xx2 = 2.D0 * quat%arr(1) * quat%arr(1)
    yy2 = 2.D0 * quat%arr(2) * quat%arr(2)  
    zz2 = 2.D0 * quat%arr(3) * quat%arr(3)
    xy2 = 2.D0 * quat%arr(1) * quat%arr(2)
    wz2 = 2.D0 * quat%arr(4) * quat%arr(3)
    zx2 = 2.D0 * quat%arr(3) * quat%arr(1)
    wy2 = 2.D0 * quat%arr(4) * quat%arr(2)
    yz2 = 2.D0 * quat%arr(2) * quat%arr(3)
    wx2 = 2.D0 * quat%arr(4) * quat%arr(1)
    
    
    quat%matrix(1,1) = 1.D0 - yy2 - zz2


    quat%matrix(1,2) = xy2 - wz2
    quat%matrix(1,3) = zx2 + wy2
    quat%matrix(2,1) = xy2 + wz2
    quat%matrix(2,2) = 1.D0 - xx2 - zz2
    quat%matrix(2,3) = yz2 - wx2
    quat%matrix(3,1) = zx2 - wy2
    quat%matrix(3,2) = yz2 + wx2
    quat%matrix(3,3) = 1.D0 - xx2 - yy2
 

    end subroutine  QcalcMatrix
    

    !-------------------------------------------------------------------------------------------

    subroutine  QfromMatrix(quat)

    !-------------------------------------------------------------------
    ! Calculate quaternion quat%arr  from  rotation matrix quat%matrix
    ! Adapted from Quaternions.py 
    !-------------------------------------------------------------------

    implicit none
    
    class (quaternion_type)  :: quat
    integer(kind=int32)      :: ierr
    real(kind=real64)        :: m(3,3)
    real(kind=real64)        :: mdiag(4)
    real(kind=real64)        :: fac  
    integer(kind=int32),dimension(1)      :: locarr    
    integer(kind=int32)      :: imax   


    ierr=0

    m = quat%matrix

    mdiag(1) = 1.D0 + m(1,1) - m(2,2) - m(3,3)
    mdiag(2) = 1.D0 - m(1,1) + m(2,2) - m(3,3)
    mdiag(3) = 1.D0 - m(1,1) - m(2,2) + m(3,3)
    mdiag(4) = 1.D0 + m(1,1) + m(2,2) + m(3,3)

    locarr = maxloc(mdiag) 
    imax = locarr(1)

    quat%arr(:) = 0.D0
    quat%arr(imax) =  0.5D0 * sqrt(mdiag(imax))
    fac = 0.25D0 /  quat%arr(imax)   

    if (imax == 1) then
      quat%arr(2) = (m(2,1) + m(1,2)) * fac 
      quat%arr(3) = (m(3,1) + m(1,3)) * fac 
      quat%arr(4) = (m(3,2) - m(2,3)) * fac 
    else if  (imax == 2) then
      quat%arr(1) = (m(2,1) + m(1,2)) * fac 
      quat%arr(3) = (m(3,2) + m(2,3)) * fac 
      quat%arr(4) = (m(1,3) - m(3,1)) * fac       
   else if  (imax == 3) then
      quat%arr(1) = (m(3,1) + m(1,3)) * fac 
      quat%arr(2) = (m(3,2) + m(2,3)) * fac 
      quat%arr(4) = (m(2,1) - m(1,2)) * fac             
   else if  (imax == 4) then  
      quat%arr(1) = (m(3,2) - m(2,3)) * fac 
      quat%arr(2) = (m(1,3) - m(3,1)) * fac 
      quat%arr(3) = (m(2,1) - m(1,2)) * fac             
   endif  
    


    end subroutine  QfromMatrix

!---------------------------------------------------------  


    subroutine  QcalcQuat(quat,ierr)

    !---------------------------------------------------------------------------
    ! Calculate  x,y,z,w array quat%arr from attitude values in  quat%att 
    ! (RA, DEC, ROLL [radians]).
    ! Adapted from
    ! http://www.euclideanspace.com/maths/geometry/rotations/conversions/eulerToQuaternion/index.htm
    ! heading -> RA
    ! attitude -> DEC
    ! bank     -> ROLL
    ! x_aircraft  ->   x_sat
    ! y_ac        ->   z_sat       
    ! z_ac        ->  -y_sat
    !---------------------------------------------------------------------------
      
    
    class (quaternion_type)     :: quat
    integer(kind=int32)         :: ierr

    real(kind=real64)           :: c1,c2,c3,s1,s2,s3

    character(len=*),parameter  :: rtname='QCalcQuat'


    ierr=0

    c1=cos(quat%att(1)/2.D0)
    c2=cos(quat%att(2)/2.D0)
    c3=cos(quat%att(3)/2.D0)
    s1=sin(quat%att(1)/2.D0)
    s2=sin(quat%att(2)/2.D0)
    s3=sin(quat%att(3)/2.D0)

    quat%arr(1) = s1*s2*c3 + c1*c2*s3
!    quat%arr(2) = -(c1*s2*c3 - s1*c2*s3)
    quat%arr(2) = s1*c2*s3 - c1*s2*c3
    quat%arr(3) = s1*c2*c3 + c1*s2*s3
    quat%arr(4) = c1*c2*c3 - s1*s2*s3

    end  subroutine  QcalcQuat

!--------------------------------------------------------------------------------

    subroutine  QcalcAtt2(quat)

    !----------------------------------------------------------------------------
    ! Calculate attitude  quat%att (heading,attitude,bank) from 
    ! x,y,z,w array quat%arr 
    ! Reference system:  X-axis: heading=0,attitude=0
    !                    Y-axis: zenith 
    !                    z-axis: aircraft axis   
    ! see 
    ! www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/
    !----------------------------------------------------------------------------------

    implicit none
    
    class (quaternion_type)     :: quat
    integer(kind=int32)         :: ierr
    real(kind=real64)           :: qw,qx,qy,qz, x_ac,y_ac,z_ac
    real(kind=real64)           :: xy_zw
    real(kind=real64),parameter :: eps = 1.D-6,dpi=3.14159265358979   
    character(len=*),parameter  :: rtname='QCALCATT2'

    ierr=0

!    write(*,*) 'QcalcAtt2'

   ! if (.not. associated(quat%att)) allocate(quat%att(3))
    quat%att(:)=0.D0
    
    qw=quat%arr(4)
    qx=quat%arr(1)  
    qy=quat%arr(2)
    qz=quat%arr(3)

    x_ac=qx
    y_ac=qz
    z_ac=-qy

    xy_zw=x_ac*y_ac + z_ac*qw


!    quat%att(1) = atan2(2*y_ac*qw-2*x_ac*z_ac , 1.D0 - 2D0*x_ac*x_ac - 2.D0*z_ac*z_ac)
!    quat%att(2) = asin(2D0*x_ac*y_ac + 2D0*z_ac*qw) 
!    quat%att(3) = atan2(2*x_ac*qw-2*y_ac*z_ac , 1.D0 - 2D0*x_ac*x_ac - 2.D0*z_ac*z_ac)
    
    if (abs(xy_zw - 0.5) < eps) then 
      ! north pole
      quat%att(1) = 2.D0 * atan2(x_ac,qw)
      quat%att(2) = dpi/2.D0 
      quat%att(3) = 0.D0
    else if  (abs(xy_zw + 0.5) < eps) then
      ! south pole
      quat%att(1) = -2.D0 * atan2(x_ac,qw)
      quat%att(2) = -dpi/2.D0 
      quat%att(3) = 0.D0      
    else
!     quat%att(1) = atan2(wy2-xz2,1.D0-yy2-zz2 )
!     quat%att(2) = asin(2.D0 * xy_zw)
!     quat%att(3) = atan2(wx2-yz2, 1.D0-xx2-zz2)

     quat%att(1) = atan2(2*y_ac*qw-2*x_ac*z_ac , 1.D0 - 2D0*x_ac*x_ac - 2.D0*z_ac*z_ac)
     quat%att(2) = asin(2.D0*x_ac*y_ac + 2.D0*z_ac*qw) 
     quat%att(3) = atan2(2*x_ac*qw-2*y_ac*z_ac , 1.D0 - 2D0*x_ac*x_ac - 2.D0*z_ac*z_ac)
 
    endif
 

    end subroutine  QcalcATT2
    
!---------------------------------------------------------  
    




!---------------------------------------------------------  



    subroutine  QcalcAtt(quat)

    !----------------------------------------------------------------------------------
    ! Calculate attitude  quat%att (ra,dec,roll) from 
    ! x,y,z,w array quat%arr 
    ! Positive x-axis:ra=0,dec=0, z-axis:north equatorial pole
    ! modified from
    ! www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/
    !----------------------------------------------------------------------------------

    implicit none
    
    class (quaternion_type)     :: quat
    integer(kind=int32)         :: ierr
    real(kind=real64)           :: xx2,yy2,zz2,wy2,xz2,xy2,wz2,wx2,yz2
    real(kind=real64)           :: xy_zw,xz_yw
    real(kind=real64),parameter :: eps = 1.D-9,dpi=3.14159265358979   
    character(len=*),parameter  :: rtname='QCALCATT'

    ierr=0

!    write(*,*) 'QcalcAtt' 

  !  if (.not. associated(quat%att)) allocate(quat%att(3))
    quat%att(:)=0.D0
   
    xx2 = 2.D0 * quat%arr(1) * quat%arr(1)
    yy2 = 2.D0 * quat%arr(2) * quat%arr(2)  
    zz2 = 2.D0 * quat%arr(3) * quat%arr(3)
    xy2 = 2.D0 * quat%arr(1) * quat%arr(2)
    xz2 = 2.D0 * quat%arr(1) * quat%arr(3)
    yz2 = 2.D0 * quat%arr(2) * quat%arr(3)
    wx2 = 2.D0 * quat%arr(4) * quat%arr(1)
    wy2 = 2.D0 * quat%arr(4) * quat%arr(2)
    wz2 = -2.D0 * quat%arr(4) * quat%arr(3)
    !xy_zw = quat%arr(1)*quat%arr(2) + quat%arr(3)*quat%arr(4)
    xz_yw=  quat%arr(1)*quat%arr(3) + quat%arr(2)*quat%arr(4)
!    write(*,*)   quat%arr(1),quat%arr(3),quat%arr(2),quat%arr(4)
!    write(*,*)  'xz_yw=',xz_yw
    if (abs(xz_yw -  0.5) < eps) then 
      ! north pole
    !  write(*,*) 'north pole'
      quat%att(1) = -2.D0 * atan2(quat%arr(1),quat%arr(4))
      quat%att(2) = asin(2.D0 * xz_yw)
      quat%att(3) = 0.D0
    else if  (abs(xz_yw + 0.5) < eps) then
      ! south pole
   !   write(*,*) 'south pole'
      quat%att(1) = -2.D0 * atan2(quat%arr(1),quat%arr(4))
      quat%att(2) = asin(2.D0 * xz_yw)
      quat%att(3) = 0.D0      
    else
!      write(*,*) 'ra_arg1=',xy2+xz2,' ra_arg2=',1.D0-zz2-yy2
!      write(*,*) 'roll_arg1=',wx2+yz2,' roll_arg2=',1.D0-xx2-yy2
      quat%att(1) = atan2(xy2+xz2,1.D0-zz2-yy2 )
      quat%att(2) = asin(2.D0 * xz_yw)
      quat%att(3) = atan2(wx2+yz2, 1.D0-xx2-yy2)
    endif
 

    end subroutine  QcalcATT
    
!---------------------------------------------------------  

    subroutine  QcalcAtt1(quat)


    !----------------------------------------------------------------------------------
    ! Calculate attitude  quat%att (ra,dec,roll) from 
    ! x,y,z,w array quat%arr 
    ! Reference point:
    ! Positive x-axis:ra=0,dec=0  z-axis:north equatorial pole
    ! 
    !-------------------------------------------------------------------------------


    implicit none
    
    class (quaternion_type)        :: quat
    integer(kind=int32)            :: ierr

    real(kind=real64),dimension(4) :: q2
    real(kind=real64)              :: xa,xb,xn,yn,zn
    real(kind=real64)              :: xy_zw,xz_yw
    real(kind=real64),parameter    :: eps = 1.D-9,dpi=3.14159265358979   
    character(len=*),parameter     :: rtname='QCALCATT'

    ierr=0

   ! if (.not. associated(quat%att)) allocate(quat%att(3))
    quat%att(:)=0.D0
      
 !   write(*,*) rtname,' quat%arr:', quat%arr

    q2 = (quat%arr) * (quat%arr)
    
    xa = q2(1) - q2(2) - q2(3) + q2(4) 
    xb = 2 * (quat%arr(1) * quat%arr(2) + quat%arr(3) * quat%arr(4)) 
    xn = 2 * (quat%arr(1) * quat%arr(3) - quat%arr(2) * quat%arr(4)) 
    yn = 2 * (quat%arr(2) * quat%arr(3) + quat%arr(1) * quat%arr(4)) 
    zn = q2(4) + q2(3) - q2(1) - q2(2) 

    quat%att(1) = atan2(xb , xa) 
    quat%att(2) = atan2(xn , sqrt(max(0.D0,1.D0 - xn**2)))
    quat%att(3) = atan2(yn , zn) 

    if  (quat%att(1) < 0.D0) quat%att(1)=quat%att(1)+2.D0*dpi
    if  (quat%att(3) < 0.D0) quat%att(3)=quat%att(3)+2.D0*dpi
    

    end subroutine QcalcAtt1



!---------------------------------------------------------  



    subroutine QfromAtt(quat)


    !-------------------------------------------------------------------------------------------------
    ! Calculate quaternion quat%arr  from attitude quat%att
    ! Converts attitude quat.att (given as RA, DEC, ROLL [radians]) to quaternion quat.arr ([x,y,z,w]) 
    ! The following assumptions are made: at att=(0,0,0) the X-axis points to  RA,DEC=(0,0),
    ! the Y-axis points at RA=+90d, DEC=0.0, the Z-Axis points at dec=+90d
    !-------------------------------------------------------------------------------------------------

      class (quaternion_type)  :: quat
      integer(kind=int32)      :: ierr
      real(kind=real64)        :: c1,c2,c3,s1,s2,s3
    

      ierr=0
    
      c1=cos(quat%att(1)/2.D0)
      c2=cos(quat%att(2)/2.D0)
      c3=cos(quat%att(3)/2.D0)
      s1=sin(quat%att(1)/2.D0)
      s2=sin(quat%att(2)/2.D0)
      s3=sin(quat%att(3)/2.D0)


      quat%arr(4) = c1*c2*c3 - s1*s2*s3
      quat%arr(1) = s1*s2*c3 + c1*c2*s3 
      quat%arr(2) = s1*c2*s3 - c1*s2*c3
      quat%arr(3) = s1*c2*c3 + c1*s2*s3    


    end subroutine QfromAtt


  end module quaternions
